class InvalidArgumentError(Exception):
    "Raised when invalid arguments are passed to class constructors and/or methods"

    def __init__(self, message):
        if message:
            super().__init__(message)


class UnavailableDataError(Exception):
    def __init__(self, message):
        if message:
            super().__init__(message)


class GenericDownloadError(Exception):
    def __init__(self, message):
        if message:
            super().__init__(message)


class UnspecifiedFontPathError(Exception):
    def __init__(self, message):
        if message:
            super().__init__(message)


class MismatchedBoundsError(Exception):
    def __init__(self, message):
        if message:
            super().__init__(message)


class MismatchedCRS(Exception):
    def __init__(self, message):
        if message:
            super().__init__(message)


class MismatchedImageSize(Exception):
    def __init__(self, message):
        if message:
            super().__init__(message)



from pathlib import Path
import goesvideo
from goesvideo.utils import gistools

"""
This demo creates a video clip of the Hawaiian Islands region using data from GOES-West
The compositor selected for this video is 'true_color_with_night_ir', which is a standard compositor defined in
the Satpy file etc/composites/abi.yaml and is also included in the GoesVideo file etc/abibands.yaml. If custom 
compositors are defined by the user in Satpy, the user should be sure to also update etc/abibands.yaml.
options in abi.yaml.
"""

#---- Custom options to be set by user
basedir = str(Path.home()) + '/goesimagery'
fontpath = str(Path.home()) + '/.local/share/fonts/meslolgldz-nerd-font/MesloLGSDZNerdFont-Bold.ttf'  # Required for timestamps
coastpath = str(Path.home()) + '/goes-workspace/coastlines'   # Required for coastlines/borders; this should be the root path
                                                              # to the GHHSG shapefiles that can be downloaded here:
                                                              # https://www.soest.hawaii.edu/pwessel/gshhg/

#---- Coords for bbox in DMS units
uldms = ('24d56m50s N', '165d08m15s W')
lrdms = ('15d19m31s N', '150d51m23s W')

#---- Convert to DD units for bbox
uldd = gistools.convert_dms_to_dd(uldms)
lrdd = gistools.convert_dms_to_dd(lrdms)
bbox = (uldd[1], lrdd[0], lrdd[1], uldd[0])

#---- Create animator
animator = goesvideo.GoesAnimator('goes-west', 'full', 'ABI-L2-CMIP',
                                  base_dir=basedir)

# Configure coastline/border options for video
coastopts = {'level_coast': 1, 'level_borders': 2}
animator.compositor.set_coastlines_options(path=coastpath, **coastopts)

# Create video
animator.create_video('true_color_with_night_ir',
                      start_time='2024-10-12 17:00:00',   #Supplied as UTC time
                      end_time='2024-10-12 23:00:00',     #Supplied as UTC time
                      bbox=bbox,
                      folder_name='Hawaii_Sample_Scene',
                      interval=1,
                      timestamps=False,
                      coastlines=True,
                      res='auto',
                      codec='rawvideo',
                      fps=20)






import os
import shutil
import sys
import tempfile
from importlib.resources import files as importfiles
from pathlib import Path

import pytest
import pytz
from PIL import Image

from goesvideo import GoesAnimator
from goesvideo.tests import treegenerator

try:
    import cv2

    has_cv2 = True
except ModuleNotFoundError:
    has_cv2 = False


def test_preview():
    # Toggle display of image
    show_image = True

    # Generate expected folder tree
    tmpfolder = tempfile.TemporaryDirectory()
    base_dir = Path(tmpfolder.name)
    treegenerator.generate_tree(base_dir, "BigHorn", copy_ncfiles=True)

    ga = GoesAnimator("goes-east", "full", "ABI-L2-CMIP", base_dir=base_dir)

    imgpath = importfiles("goesvideo") / "tests" / "Test Images"
    imgfiles = list(imgpath.glob("*.png"))

    fontpath = str(importfiles("goesvideo") / "tests" / "Fonts" / "Roboto-Regular.ttf")

    timestamps = {
        "label": "2023-01-01 00:00:00",
        "position": "upper-left",
        "fontpath": fontpath,
        "fontcolor": (255, 0, 0),
        "tzinfo": (pytz.utc, "UTC"),
        "fontsize": 20,
        "opacity": 0.7,
    }

    text = {
        "label": "this is a test label",
        "position": (600, 600),
        "fontpath": fontpath,
        "fontcolor": (255, 0, 0),
        "fontsize": 20,
        "opacity": 0.7,
    }

    if has_cv2:
        arrow = {
            "label": {
                "label": "this is an arrow",
                "padding": (10, 10),
                "fontpath": fontpath,
                "fontcolor": (255, 0, 0),
                "fontsize": 20,
                "opacity": 0.7,
            },
            "tiplength": 0.1,
            "width": 2,
            "color": (255, 0, 0),
            "opacity": 0.7,
            "start_position": (200, 200),
            "end_position": (400, 400),
        }

    circle = {
        "label": {
            "label": "this is a circle",
            "padding": (10, 10),
            "fontpath": fontpath,
            "fontcolor": (255, 0, 0),
            "fontsize": 20,
            "opacity": 0.3,
        },
        "fill": (255, 0, 0),
        "outline": (0, 0, 0),
        "width": 1,
        "radius": 5,
        "centerpos": (500, 500),
    }

    kwargs = {}
    kwargs["timestamps"] = timestamps
    kwargs["text"] = text
    if has_cv2:
        kwargs["arrow"] = arrow
    kwargs["circle"] = circle

    testimg = Image.open(imgfiles[0])
    testimg = testimg.resize((1024, 768))
    tmpimgfile = tempfile.NamedTemporaryFile("w+b", suffix=".png", delete=False)
    testimg.save(tmpimgfile.name)
    testimg.close()
    img = ga.preview(
        use_image_file=tmpimgfile.name, **kwargs, res=(1024, 768), display=False
    )

    assert isinstance(img, Image.Image)

    if show_image:
        img.show()

    img.close()

    # Cleanup
    try:
        tmpimgfile.close()
        os.unlink(tmpimgfile.name)
        shutil.rmtree(base_dir)
    except FileNotFoundError:
        pass
    except PermissionError:
        pass

    return


if __name__ == "__main__":
    sys.exit(pytest.main())



import shutil
import copy
import json
from datetime import datetime
from pathlib import Path

import matplotlib.colors as colors
import numpy as np
import pytz
from PIL import ImageDraw, ImageFont, Image
from colorama import Fore
from matplotlib import colormaps
from moviepy.editor import VideoFileClip
from moviepy.video.fx import freeze as frz

try:
    import cv2
except ModuleNotFoundError:
    print(
        f"{Fore.RED}WARNING - OpenCV package not detected. 'add_arrow' function will be unavailable."
    )


# ------ Image Annotations
def build_cmap(fname):
    """
    Builds a matplotlib cmap for a colormap specification generated using the app available at
    sciviscolor.org
    @param fname: (str) path to colormap json
    @return: (matplotlib.cm obj) colormap
    """
    try:
        with open(fname) as jsonfile:
            _json = json.load(jsonfile)
        points = _json["colormaps"][0]["points"]
        clist = []
        for p in points:
            c = (p["x"], (p["r"], p["g"], p["b"]))
            clist.append(c)

        cmap = colors.LinearSegmentedColormap.from_list("new", clist, N=256)
        cmap = cmap.reversed()
    except FileNotFoundError:
        cmap = colormaps.get_cmap(fname)

    return cmap


def copy_image(file, dstpath):
    """
    Copy an image to a new location
    @param file: (str) path to image file
    @param dstpath: (str) save path
    @return: (str) full save path to image
    """
    fp = Path(file)
    fname = dstpath / (fp.stem + fp.suffix)
    shutil.copy(file, str(fname))
    return fname


def add_text(img, **kwargs):
    """
    Add custom text to an image frame
    @param img: (PIL.Image) input image
    @param kwargs: (dict) see 'modify_image' function for options
    @return: (PIL.Image) output image
    """
    position = kwargs.get("position", "upper-left")
    fontcolor = kwargs.get("fontcolor", (0, 0, 0))
    fontpath = kwargs["fontpath"]
    fontsize = kwargs.get("fontsize", None)
    label = kwargs["label"]
    opacity = kwargs.get("opacity", 1)
    rotation = kwargs.get("rotation", 0)

    # Convert input image to RGB
    img = img.convert("RGB")

    # If fontsize not provided, attempt to scale to image resolution
    if not fontsize:
        fontsize = int(0.02 * img.height)

    font = ImageFont.truetype(fontpath, fontsize)

    # If opacity equals 1 then text can be added directly to the input image
    # Otherwise, need to create a new image layer for alpha compositing

    fontcolor = (fontcolor[0], fontcolor[1], fontcolor[2], int(opacity * 255))
    layercolor = (fontcolor[0], fontcolor[1], fontcolor[2], 0)
    txtimg = Image.new("RGBA", img.size, layercolor)
    draw = ImageDraw.Draw(txtimg)

    upper = 0
    lower = img.height
    left = 0
    right = img.width
    woffsetfrac = 0.01
    hoffsetfrac = 0.03
    txtlength = draw.textlength(label, font=font, font_size=fontsize)

    if position == "upper-left":
        x = left + int(woffsetfrac * img.width)
        y = upper + int(hoffsetfrac * img.height + fontsize)
    elif position == "upper-center":
        x = int((left + right) / 2.0)
        y = upper + int(hoffsetfrac * img.height + fontsize)
    elif position == "upper-right":
        x = right - int(woffsetfrac * img.width + txtlength)
        y = upper + int(hoffsetfrac * img.height + fontsize)
    elif position == "lower-left":
        x = left + int(woffsetfrac * img.width)
        y = lower - int(hoffsetfrac * img.height + fontsize)
    elif position == "lower-center":
        x = int((left + right) / 2.0)
        y = lower - int(hoffsetfrac * img.height + fontsize)
    elif position == "lower-right":
        x = right - int(woffsetfrac * img.width + txtlength)
        y = lower - int(hoffsetfrac * img.height + fontsize)
    else:
        x = position[0]
        y = position[1]

    draw.text((0, 0), label, fill=fontcolor, font=font, font_size=fontsize, **kwargs)
    img = img.convert("RGBA")

    # Get bounding box for text and crop to size before rotating
    bbox = draw.textbbox((0, 0), label, font=font, font_size=fontsize)
    txtimg = txtimg.crop(bbox)
    txtimg = txtimg.rotate(-rotation, expand=True)

    img.paste(txtimg, (int(x), int(y)), txtimg)

    img = img.convert("RGB")
    return img


def add_timestamps(img, label, **kwargs):
    """
    Add timestamp to image
    @param img: (PIL.Image) input image
    @param label: (str) datetime string to be added
    @param kwargs: (dict) see 'modify_image' function for options
    @return: (PIL.Image) output image
    """
    tzinfo = None
    if kwargs:
        tzinfo = kwargs.pop("timezone", None)  # tuple

    if tzinfo:
        _tz = tzinfo[0]
        _abbr = tzinfo[1]
        if not isinstance(label, datetime):
            _t = datetime.fromisoformat(label)
        else:
            _t = label
        try:
            tstamp = pytz.utc.localize(_t).astimezone(_tz)
        except ValueError:
            tstamp = _t.astimezone(_tz)
        label = tstamp.isoformat()[0:-6].replace("T", " ") + " " + _abbr
    else:
        label = label.replace("T", " ") + " " + "UTC"

    if kwargs:
        kwargs["label"] = label
    else:
        kwargs = {}
        kwargs["label"] = label

    img = add_text(img, **kwargs)

    return img


def add_circle(img, centerpos, radius, label=None, fill=None, outline=None, width=None):
    """
    Add a circle to an image
    @param img: (PIL.Image) input image
    @param centerpos: (tup) center position of circle in pixels (w, h)
    @param radius: (int) radius of circle in pixels
    @param label: (dict) label dictionary
    @param fill: (str) color to use for filling circle
    @param outline: (str) color to use for circle outline
    @param width: (int) line width in pixels
    @return: (PIL.Image) output image
    """

    # Unpack label dict
    if label:
        padding = label.get("padding", (20, 20))
        txt = label["label"]
        fontpath = label["fontpath"]
        fontsize = label.get("fontsize", 20)
        font = ImageFont.truetype(fontpath, fontsize)

    draw = ImageDraw.Draw(img)

    # Convert coords to bounding box
    xcenter = centerpos[0]
    ycenter = centerpos[1]
    x0 = xcenter - radius
    x1 = xcenter + radius
    y0 = ycenter - radius
    y1 = ycenter + radius
    xy = [(x0, y0), (x1, y1)]

    # Add circle
    draw.ellipse(xy, fill=fill, outline=outline, width=width)

    # Add anchor text if label has been provided
    if label:
        txtlength = draw.textlength(txt, font=font, font_size=fontsize)
        _x = xcenter + padding[0]
        if _x + txtlength > img.width:
            _x = _x - txtlength
        _y = ycenter - padding[1]

        kwargs = label
        kwargs.pop("padding")
        kwargs["position"] = (_x, _y)

        img = add_text(img, **kwargs)

    return img


def add_arrow(
    img,
    startpos,
    endpos,
    label=None,
    tiplength=0.1,
    width=8,
    color=(255, 0, 0),
    **kwargs,
):
    """
    Add an arrow to the image
    @param img: (PIL.Image obj) input image
    @param startpos: (tup) start position of arrow in pixels (w, h)
    @param endpos: (tup) end position of arrow in pixels (w, h)
    @param label: (dict) label dictionary
    @param tiplength: (int) length of arrow tip in pixels
    @param width: (int) width of line in pixels
    @param color: (tup) BGR color
    @param kwargs: (dict) see 'modify_image' function for options
    @return: (PIL.Image obj) output image
    """

    # Unpack label dict if provided
    if label:
        padding = label.get("padding", (20, 20))
        txt = label["label"]
        fontpath = label["fontpath"]
        opacity = label.get("opacity", 1)
        fontcolor = label.get("fontcolor", (0, 0, 0))
        fontsize = label.get("fontsize", 20)
        font = ImageFont.truetype(fontpath, fontsize)
        draw = ImageDraw.Draw(img)

    # Unpack arrow opacity
    if kwargs:
        arrow_opacity = kwargs.get("opacity", 1)
    else:
        arrow_opacity = 1

    # If opacity less than 1, need to create new image layer for alpha compositing
    if arrow_opacity != 1:
        img = img.convert("RGB")
        arrowimg = img
        img = img.convert("RGBA")
        imgarr = np.array(arrowimg)
        imgarr = cv2.arrowedLine(
            imgarr, startpos, endpos, color, thickness=width, tipLength=tiplength
        )
        arrowimg = Image.fromarray(imgarr)
        arrowimg = arrowimg.convert("RGBA")
        arrowimg.putalpha(int(arrow_opacity * 255))
        img = Image.alpha_composite(img, arrowimg)
    else:
        img = img.convert("RGB")
        imgarr = np.array(img)
        imgarr = cv2.arrowedLine(
            imgarr, startpos, endpos, color, thickness=width, tipLength=tiplength
        )
        img = Image.fromarray(imgarr)

    # Add label if provided
    if label:
        # Calculate position of label using start position of arrow
        label_x = 0
        label_y = 0
        # Put the text to the left if arrow points from left to right
        if startpos[0] <= endpos[0]:
            txtlength = int(draw.textlength(txt, font=font, font_size=fontsize))
            _x = startpos[0] - (padding[0] + txtlength)
            if _x < 0:
                _x = startpos[0] + padding[0]

            # Put the text below the arrow if points from bottom to top
            if startpos[1] <= endpos[1]:
                _y = startpos[1] - padding[1]
                if _y > img.height:
                    _y = startpos[1]

            # Put the text above the arrow if points from top to bottom
            else:
                _y = startpos[1] + padding[1]
                if _y < 0:
                    _y = 0

            label_x = _x
            label_y = _y

        # Put the text to the right if arrow points from right to left
        else:
            txtlength = int(draw.textlength(txt, font=font, font_size=fontsize))
            _x = startpos[0] + padding[0]
            if _x + txtlength > img.width:
                _x = img.width - txtlength

            # Put the text below the arrow if points from bottom to top
            if startpos[1] <= endpos[1]:
                _y = startpos[1] - padding[1]
                if _y > img.height:
                    _y = startpos[1]

            # Put the text above the arrow if points from top to bottom
            else:
                _y = startpos[1] + padding[1]
                if _y < 0:
                    _y = 0

            label_x = _x
            label_y = _y

        position = (label_x, label_y)

        # Build text dict
        textdict = {
            "position": position,
            "label": txt,
            "fontpath": fontpath,
            "fontcolor": fontcolor,
            "fontsize": fontsize,
            "opacity": opacity,
        }

        kwargs = {**textdict, **kwargs}

        # Add text to image
        img = add_text(img, **kwargs)

    return img


def add_triangle(img, xy, fill=None, outline=None, width=1):
    """
    Add a triangle to an image
    @param img: (PIL.Image obj) input image
    @param xy: (tup) location of triangle in pixels (w, h)
    @param fill: (str) color to use for filling triangle
    @param outline: (str) color to use for outline
    @param width: (int) width of outline in pixels
    @return: (PIL.Image obj) output image
    """
    draw = ImageDraw.Draw(img)
    draw.polygon(xy, fill=fill, outline=outline, width=width)

    return img


def convert_color(img):
    sub = img.convert("RGBA")
    data = np.array(sub)
    red, green, blue, alpha = data.T
    data = np.array([blue, green, red, alpha])
    data = data.transpose()
    sub = Image.fromarray(data)

    return sub


def modify_image(img, **kwargs):
    """
    Helper function to perform image modifications

    @param img: (PIL.Image obj) image to be modified
    @param kwargs: (dict) Valid keyword arguments include:
                   - 'res': (tup) output resolution w, h in pixels
                   - 'cmap': (str) or (matplotlib.colormap) applies a colormap to the image
                             If string is provided, it must either be a name of an existing
                             matplotlib colormap or a path to a json containing a colormap
                             generated by the online app at https://sciviscolor.org/color-moves-app/
                   - 'timestamps': (dict) adds a timestamp to the image
                                   {'label': (str) timestamp text,
                                    'tdict': {'position': (tup) of x,y pixel coords e.g. (1000, 200) OR
                                                          (str) location string e.g. 'upper-left',
                                              'fontpath': (str) path to ttf font,
                                              'fontcolor': (tup) RGB color of font,
                                              'fontsize': (int) size of font in pixels,
                                              'tzinfo': (tup) of pytz timezone object and 3-char tz
                                                        abbr.
                                              }
                                    }
                   - 'text': (dict) adds custom test to the image
                                    {'label': (str) text to add,
                                     'position': (tup) of x,y pixel coords e.g. (1000, 200)
                                     'fontpath': (str) path to ttf font
                                     'fontcolor': (tup) RGB color of font,
                                     'fontsize': (int) size of font in pixels
                                     'rotation': (float) angle in degrees for clockwise rotation
                                     }
                   - 'arrow': (dict) add an arrow to the image, and optionally, an anchored text
                                     string
                                     {'start_positon': (tup) x, y pixel coord of start position
                                      'end_position': (tup) x, y pixel coord of end position
                                      'tiplength': (float) length of arrow tip as fraction of arrow length
                                      'width': (int) width of arrow in pixels
                                      'color': (tup) BGR color
                                      'label': (dict) {'label': (str) text string to display,
                                                       'padding': (tup) padding between label and arrow start in pixels
                                                       'fontpath': (str) path to ttf font,
                                                       'fontcolor': (tup) RGB color,
                                                       'fontsize': (int) font size in pixels
                                                       },
                                      }
                   - 'circle': (dict) add a circle to the image, and optionally, an anchored text
                                string
                                      {'centerpos': (tup) x, y pixel coord of center position
                                       'radius': (int) radius of circle in pixels,
                                       'label': (dict) {'label': (str) text string to display,
                                                        'padding': (tup) padding between label and
                                                                   circle in pixels (x, y),
                                                        'fontpath': (str) path to ttf font,
                                                        'fontcolor': (tup) RGB color,
                                                        'fontsize': (int) font size in pixels
                                                        },
                                       'fill': (tup) RGB fill color for circle,
                                       'outline': (tup) RGB color of circle outline,
                                       'width': (int) width of circle outline in pixels,
                                       }
                   - 'triangle': (dict) {'coords': (list) x, y pixel coordinates of vertices as a tuple
                                         'fill': (tup) RGB fill color for triangle,
                                         'outline': (tup) RGB color of triangle outline,
                                         'width': (int) width of triangle outline in pixels

    @return: PIL Image
    """
    kwargs_copy = copy.deepcopy(kwargs)
    # Unpack kwargs
    cmap = kwargs.get("cmap", None)
    timestamps = kwargs.get("timestamps", None)
    text = kwargs.get("text", None)
    res = kwargs.get("res", None)
    arrow = kwargs.get("arrow", None)
    circle = kwargs_copy.get("circle", None)
    triangle = kwargs.get("triangle", None)
    if cmap:
        if ".json" in cmap:
            cmap = build_cmap(cmap)
        else:
            cmap = colormaps.get_cmap(cmap)

        img = img.convert("L")
        imgarr = np.array(img) / 255
        img = Image.fromarray(np.uint8(cmap(imgarr) * 255))

    if timestamps:
        ftime = timestamps.get("label")
        tdict = timestamps.get("tdict")

        img = add_timestamps(img, ftime, **tdict)

    if text:
        text_new = copy.deepcopy(text)
        if isinstance(text, list):
            for t in text:
                text_new = copy.deepcopy(t)
                img = add_text(img, **text_new)
        else:
            text_new = copy.deepcopy(text)
            img = add_text(img, **text_new)

    if arrow:
        if isinstance(arrow, list):
            for a in arrow:
                arrow_new = copy.deepcopy(a)

                startpos = arrow_new.pop("start_position")
                endpos = arrow_new.pop("end_position")

                img = add_arrow(img, startpos, endpos, **arrow_new)
        else:
            arrow_new = copy.deepcopy(arrow)

            startpos = arrow_new.pop("start_position")
            endpos = arrow_new.pop("end_position")

            img = add_arrow(img, startpos, endpos, **arrow_new)

    if circle:
        if isinstance(circle, list):
            for c in circle:
                circle_new = copy.deepcopy(c)
                radius = circle_new.pop("radius", None)

                centerpos = circle_new.pop("centerpos", None)
                img = add_circle(img, centerpos, radius, **circle_new)
        else:
            circle_new = copy.deepcopy(circle)

            centerpos = circle_new.pop("centerpos", None)
            radius = circle_new.pop("radius", None)

            img = add_circle(img, centerpos, radius, **circle_new)

    if res:
        img = img.resize((res[0], res[1]))

    if triangle:
        xy = triangle.pop("coords", None)
        fill = triangle.pop("fill", None)
        outline = triangle.pop("outline", None)
        width = triangle.pop("width", 1)
        img = add_triangle(img, xy, fill=fill, outline=outline, width=width)

    kwargs = kwargs_copy
    img.convert("RGB")
    return img


# ------ Video annotations
class GoesClip(VideoFileClip):
    def __init__(self, filepath):
        super().__init__(filepath)

    def preview(self, tmark=None, display=True, **kwargs):
        """
        Grabs a sample frame from the clip and applies desired annotations

        :param tmark: (float or tup or str) time to grab frame; if float it should represent
                                            the time mark in seconds; if tup it should be in the
                                            form (hours, mins, secs); if string it should be given
                                            as 'hh:mm:ss.ms'
        :param kwargs: annotation keywords
        :return: PIL Image
        """

        # Grab frame
        if tmark:
            nparr = self.get_frame(tmark)
        else:
            _t = np.random.Generator.integers(0, high=self.duration)
            nparr = self.get_frame(_t)

        # Convert to image
        img = Image.fromarray(nparr)

        # Apply modifications
        img = modify_image(img, **kwargs)
        if display:
            img.show()

        return img

    def annotate(
        self,
        t_edit_start=None,
        t_edit_end=None,
        freeze=False,
        freeze_img=None,
        codec=None,
        **kwargs,
    ):
        """
        Add annotations to an existing animation. Modified video will be saved to the same
        path.
        :param filepath: (str) path to video file
        :param svname: (str) name of output video; if None then name will be the same
                             as the input file suffixed by '_annotated'
        :param t_edit_start: (float or tup or str) start time to apply annotation; if float it should represent
                                            the time mark in seconds; if tup it should be in the form
                                            (hours, mins, secs); if string it should be given as
                                            'hh:mm:ss.ms'
        :param t_edit_end: (float or tup or str) end time to apply annotation; if float it should represent
                                            the time mark in seconds; if tup it should be in the form
                                            (hours, mins, secs); if string it should be given as
                                            'hh:mm:ss.ms'
        :param freeze: (bool) If true, the video will pause while annotations are displayed
        :param freeze_img: (PIL Image) If provided, this image will be displayed while the video is frozen
        :param kwargs: annotation-related kwargs
        :return: None
        """

        if t_edit_start and t_edit_end:
            tstart = t_edit_start
            tend = t_edit_end
        elif t_edit_end:
            tstart = 0
            tend = t_edit_end
        elif t_edit_start:
            tstart = t_edit_start
            tend = self.duration
        else:
            tstart = 0
            tend = self.duration

        if not freeze:
            clip = self.fl(
                lambda get_frame, t: (
                    get_frame(t)
                    if tstart >= t or t >= tend
                    else np.array(
                        modify_image(Image.fromarray(get_frame(t)), **kwargs),
                        dtype=np.uint8,
                    )
                )
            )
        else:
            if not freeze_img:
                img = Image.fromarray(self.get_frame(tstart))
                if kwargs:
                    img = modify_image(img, **kwargs)
                nparr = np.array(img, dtype=np.uint8)
                frzclip = frz.freeze(
                    self, t=t_edit_start, freeze_duration=t_edit_end - t_edit_start
                )
                clip = frzclip.fl(
                    lambda get_frame, t: nparr if tstart <= t <= tend else get_frame(t)
                )
            else:
                if (
                    self.size[0] != freeze_img.size[0]
                    or self.size[1] != freeze_img.size[1]
                ):
                    kwargs["res"] = (self.size[0], self.size[1])

                if kwargs:
                    freeze_img = modify_image(freeze_img, **kwargs)

                freeze_img = freeze_img.convert("RGB")
                nparr = np.array(freeze_img, dtype=np.uint8)
                frzclip = frz.freeze(
                    self, t=t_edit_start, freeze_duration=t_edit_end - t_edit_start
                )
                clip = frzclip.fl(
                    lambda get_frame, t: nparr if tstart <= t <= tend else get_frame(t)
                )

        if codec == "rawvideo":
            retclip = clip.fl(lambda get_frame, t: np.array(convert_color(img)))
        else:
            retclip = clip

        return retclip



import shutil
from importlib.resources import files as importfiles
from pathlib import Path


def generate_tree(
    base_dir, scenename, overlay_scene=None, copy_images=False, copy_ncfiles=False
):
    p = importfiles("goesvideo") / "tests" / "Test Images"
    p2 = importfiles("goesvideo") / "tests" / "Test NC Files"

    png_fnames = list(p.glob("*.png"))
    tiff_fnames = list(p.glob("*.tif"))

    scene_dir_base = Path(base_dir / "Scenes" / scenename)
    scene_edit_dir = Path(base_dir / "SceneEdits")
    scene_edit_dir_base = scene_edit_dir / scenename
    vid_dir = Path(base_dir / "Videos" / scenename)

    if overlay_scene:
        scene_dir_overlay = Path(base_dir / "Scenes" / overlay_scene)
        scene_edit_dir_overlay = scene_edit_dir / overlay_scene

    base_dir.mkdir(parents=True, exist_ok=True)
    scene_dir_base.mkdir(parents=True, exist_ok=True)
    scene_edit_dir.mkdir(parents=True, exist_ok=True)
    scene_edit_dir_base.mkdir(parents=True, exist_ok=True)
    vid_dir.mkdir(parents=True, exist_ok=True)

    if overlay_scene:
        scene_dir_overlay.mkdir(parents=True, exist_ok=True)
        scene_edit_dir_overlay.mkdir(parents=True, exist_ok=True)

    # Copy image files to temp folders
    if copy_images:
        for f in png_fnames:
            shutil.copyfile(str(f), str(scene_dir_base / f.name))
        if overlay_scene:
            for f in tiff_fnames:
                shutil.copyfile(str(f), str(scene_dir_overlay / f.name))

        # Copy metadata.json and timestamps.csv
        shutil.copyfile(str(p / "metadata.json"), str(scene_dir_base / "metadata.json"))
        shutil.copyfile(
            str(p / "timestamps.csv"), str(scene_dir_base / "timestamps.csv")
        )
        if overlay_scene:
            shutil.copyfile(
                str(p / "metadata.json"), str(scene_dir_overlay / "metadata.json")
            )
            shutil.copyfile(
                str(p / "timestamps_overlay.csv"),
                str(scene_dir_overlay / "timestamps.csv"),
            )

    if copy_ncfiles:
        nc_path = Path(base_dir / "goesdata")
        nc_path.mkdir(parents=True, exist_ok=True)
        nc_fnames = list(p2.glob("*.nc"))
        for f in nc_fnames:
            shutil.copyfile(str(f), str(nc_path / f.name))

    return



import json
import os
import shutil
import sys
import tempfile
import time
from importlib.resources import files as importfiles
from pathlib import Path

import pytest
import pytz
import rasterio.crs
from PIL import Image

from goesvideo.addons.sceneeditor import GoesSceneEditor
from goesvideo.tests import treegenerator
from goesvideo.utils import gistools


def check_metadata(key, metadata):
    crs = rasterio.crs.CRS.from_wkt(metadata["geodata"]["crs"])
    retval = (None, None)
    if key == "bbox":
        bbox = metadata["Crop_Box"]
        # if crs != "EPSG:4326":
        #    bbox = gistools.transform_bbox(bbox, crs, "EPSG:4326")
        retval = (bbox,)
    elif key == "size":
        width = metadata["geodata"]["width"]
        height = metadata["geodata"]["height"]
        retval = (width, height)
    elif key == "crs":
        retval = (crs,)

    return retval


def get_metadata(_dir):
    with open(str(_dir / "metadata.json"), "r") as f:
        metadata = json.load(f)

    return metadata


def test_scene_editor():
    """
    This test evaluates the functions in class sceneditor.GoesSceneEditor using the png and tif images located in
    tests/Test Images. The scene depicted in these images is GOES-East, Ch02 (png) and fire detection (tif) for
    Big Horn, WY on October 8, 2024, which is when a large wildfire began in that area. The test evaluates functions
    for resizing, recropping, reprojecting, adding timestamps and adding various annotations. At the end of these tests,
    a video is generated which shows the overlay of the fire detection product on the base GOES imagery of the area.
    """

    # Toggle display of output images; if true, this will only generate and display a single preview image for each
    # test function below; if false, all 8 png and 8 tif images in the tests/Test Images folder will be manipulated
    preview = False

    # Specify base dir if you wish to inspect the output of the test functions below
    base_dir = None

    # Set fontpath
    fontpath = str(importfiles("goesvideo") / "tests" / "Fonts" / "Roboto-Regular.ttf")

    # Grab test images - png base, tif overlay
    p = importfiles("goesvideo") / "tests" / "Test Images"

    png_fnames = list(p.glob("*.png"))
    png_count = len(png_fnames)
    tiff_fnames = list(p.glob("*.tif"))
    tiff_count = len(tiff_fnames)

    # Create standard folder structure using temp dir as base dir
    if base_dir:
        base_dir = base_dir
        delete_flag = False
    else:
        base_dir_tmp = tempfile.TemporaryDirectory()
        base_dir = Path(base_dir_tmp.name)
        delete_flag = True

    treegenerator.generate_tree(
        base_dir, "BigHorn", overlay_scene="BigHornFire", copy_images=True
    )
    scene_dir_base = Path(base_dir / "Scenes" / "BigHorn")
    scene_dir_overlay = Path(base_dir / "Scenes" / "BigHornFire")
    scene_edit_dir = Path(base_dir / "SceneEdits")
    scene_edit_dir_base = scene_edit_dir / "BigHorn"
    scene_edit_dir_overlay = scene_edit_dir / "BigHornFire"

    # Initialize scene editors
    editor = GoesSceneEditor(str(base_dir), "BigHorn", session_name="BigHorn_Base_Edit")
    oleditor = GoesSceneEditor(
        str(base_dir), "BigHornFire", session_name="BigHorn_Fire_Edit"
    )
    editor.set_font(fontpath=fontpath, fontsize=20, fontcolor=(255, 0, 0))

    metadata_base = get_metadata(scene_dir_base)
    metadata_overlay = get_metadata(scene_dir_overlay)

    assert (
        len(list((scene_edit_dir_base / "BigHorn_Base_Edit").glob("*.png")))
        == png_count
    )
    assert (
        len(list((scene_edit_dir_overlay / "BigHorn_Fire_Edit").glob("*.tif")))
        == tiff_count
    )

    # Recrop images
    # original bbox: [-108.09, 43.794, -106.43, 45.25]
    new_bbox = (-108.088, 43.796, -106.432, 45.23)
    editor.recrop(new_bbox, preview=preview)
    oleditor.recrop(new_bbox, preview=preview)
    if not preview:
        metadata_base = get_metadata(scene_edit_dir_base / "BigHorn_Base_Edit")
        metadata_overlay = get_metadata(scene_edit_dir_overlay / "BigHorn_Fire_Edit")
        assert (
            len(list((scene_edit_dir_base / "BigHorn_Base_Edit").glob("*.png")))
            == png_count
        )
        assert (
            len(list((scene_edit_dir_overlay / "BigHorn_Fire_Edit").glob("*.tif")))
            == tiff_count
        )
        assert gistools.check_bbox(
            new_bbox, check_metadata("bbox", metadata_base)[0], degtol=0.5
        ) == (True, True)
        assert gistools.check_bbox(
            new_bbox, check_metadata("bbox", metadata_overlay)[0], degtol=0.5
        ) == (True, True)

    # Resize images
    # original size: 322 x 207
    w, h = (1024, 768)
    editor.resize(w, h, preview=preview)
    oleditor.resize(w, h, preview=preview)
    if not preview:
        metadata_base = get_metadata(scene_edit_dir_base / "BigHorn_Base_Edit")
        metadata_overlay = get_metadata(scene_edit_dir_overlay / "BigHorn_Fire_Edit")
        assert (
            len(list((scene_edit_dir_base / "BigHorn_Base_Edit").glob("*.png")))
            == png_count
        )
        assert (
            len(list((scene_edit_dir_overlay / "BigHorn_Fire_Edit").glob("*.tif")))
            == tiff_count
        )
        assert check_metadata("size", metadata_base) == (w, h)
        assert check_metadata("size", metadata_overlay) == (w, h)

    # Reproject images
    # original projection: Geostationary
    crs = "EPSG:4326"
    editor.reproject(rasterio.crs.CRS.from_epsg(4326), preview=preview)
    oleditor.reproject(rasterio.crs.CRS.from_epsg(4326), preview=preview)
    if not preview:
        metadata_base = get_metadata(scene_edit_dir_base / "BigHorn_Base_Edit")
        metadata_overlay = get_metadata(scene_edit_dir_overlay / "BigHorn_Fire_Edit")
        assert (
            len(list((scene_edit_dir_base / "BigHorn_Base_Edit").glob("*.png")))
            == png_count
        )
        assert (
            len(list((scene_edit_dir_overlay / "BigHorn_Fire_Edit").glob("*.tif")))
            == tiff_count
        )
        assert check_metadata("crs", metadata_base)[0] == crs
        assert check_metadata("crs", metadata_overlay)[0] == crs

    # Add timestamps
    editor.add_timestamps(
        "upper-left", pytz.timezone("US/Mountain"), "MDT", preview=preview
    )
    if not preview:
        assert (
            len(list((scene_edit_dir_base / "BigHorn_Base_Edit").glob("*.png")))
            == png_count
        )
        assert (
            len(list((scene_edit_dir_overlay / "BigHorn_Fire_Edit").glob("*.tif")))
            == tiff_count
        )

    # Annotate images
    editor.add_annotation(
        "text",
        ("45d04m17s N", "106d48m40s W"),
        label="MT",
        labelopts={"fontsize": 20, "rotation": 0},
    )
    editor.add_annotation(
        "text",
        ("44d57m36s N", "106d48m40s W"),
        label="WY",
        labelopts={"fontsize": 20, "rotation": 0},
    )

    circleopts = {"radius": 7, "fill": (0, 0, 255), "outline": (0, 0, 0), "width": 1}

    editor.add_annotation(
        "circle",
        ("44d21m56s N", "107d10m28s W"),
        label="3504m elev.",
        labelopts={"fontsize": 15, "padding": (15, 15)},
        **circleopts,
    )

    editor.process_annotations(preview=preview)
    if not preview:
        assert (
            len(list((scene_edit_dir_base / "BigHorn_Base_Edit").glob("*.png")))
            == png_count
        )
        assert (
            len(list((scene_edit_dir_overlay / "BigHorn_Fire_Edit").glob("*.tif")))
            == tiff_count
        )

    # Create overlay
    editor.add_overlay(
        oleditor, overlay_session_name="Big_Horn_Overlay", preview=preview
    )
    overlay_path = base_dir / "SceneEdits" / "BigHorn" / "Big_Horn_Overlay"
    assert len(list(overlay_path.glob("*.png"))) == png_count
    if preview:
        for f in overlay_path.glob("*.png"):
            img = Image.open(f)
            img.show()
            time.sleep(0.3)

    # Create video
    neweditor = GoesSceneEditor(
        str(base_dir), scene_folder="BigHorn", session_name="Big_Horn_Overlay"
    )
    neweditor.to_video(codec="mpeg4", fps=20)
    vid_path = base_dir / "Videos" / "BigHorn" / "Big_Horn_Overlay" / "video.mp4"
    assert os.path.exists(str(vid_path))

    # Cleanup
    if delete_flag:
        try:
            shutil.rmtree(base_dir)
        except FileNotFoundError:
            pass


if __name__ == "__main__":
    sys.exit(pytest.main())



import sys
import tempfile
from pathlib import Path

import pytest

from goesvideo import GoesAnimator
from goesvideo.tests import treegenerator


def test_animating():
    # Create base dir tree
    tmppath = tempfile.TemporaryDirectory()
    base_dir = Path(tmppath.name)
    treegenerator.generate_tree(base_dir, "C02", copy_images=True)

    # Create videos
    exists = []
    ga = GoesAnimator("goes-east", "conus", "ABI-L2-CMIP", base_dir=str(base_dir))
    ga.create_video("C02", from_existing_imgs=True, fps=1, force=True)
    exists.append((base_dir / "Videos" / "C02").exists())

    ga = GoesAnimator("goes-east", "conus", "ABI-L2-CMIP", base_dir=str(base_dir))
    ga.create_video("C02", from_existing_imgs=True, fps=1, cmap="Spectral", force=True)
    exists.append((base_dir / "Videos" / "C02").exists())

    ga = GoesAnimator("goes-east", "conus", "ABI-L2-CMIP", base_dir=str(base_dir))
    ga.create_video("C02", from_existing_imgs=True, fps=1, force=True)
    exists.append((base_dir / "Videos" / "C02").exists())

    assert all(exists)

    return


if __name__ == "__main__":
    sys.exit(pytest.main())



import copy
import os
import tempfile
from datetime import datetime
from pathlib import Path

import numpy as np
import pytz
import rasterio
import rasterio.crs
import rasterio.warp as rwarp
import rasterio.windows as rwindows
from PIL import Image
from colorama import Fore
from rasterio.coords import BoundingBox
from rasterio.enums import Resampling
from rasterio.transform import from_bounds, rowcol
from shapely.geometry import box

"""
Functions for manipulating tif and png georeferenced images
"""

# ------ Coordinate calculations and conversions


def convert_dms_to_dd(coords):
    """
    Convert Degrees-Minutes-Seconds coords to decimal degrees
    :param coords: (tup) (lat, lon) e.g. ('42d30m22s N',
    :param lon: (str) e.g. '87d15m50s W'
    :return: (tup) (lat, lon)
    """
    dd_ret_list = []
    for coord in coords:
        deg = int(coord.split("d")[0])
        minutes = int((coord.split("d")[1]).split("m")[0])
        secs = int((coord.split("m")[1]).split("s")[0])
        direction = coord[-1]
        dd = deg + (minutes / 60.0) + (secs / 3600)
        if direction == "W" or direction == "S":
            dd = -dd
        dd_ret_list.append(dd)

    dd_ret = tuple(dd_ret_list)

    return dd_ret


def convert_point(point, point_crs, geodata=None):
    geodata_copy = copy.deepcopy(geodata)
    dst_crs = geodata_copy["crs"]

    pointx, pointy = rwarp.transform(point_crs, dst_crs, [point[1]], [point[0]])

    return pointy[0], pointx[0]


def xy_from_latlon(point, geodata=None):
    """
    Transform lat/lon coordinates to pixel coordinates
    :param point: (tup) (lat, lon)
    :param geodata: (dict) {'bbox': (tup) (west, south, east, north) in image crs,
                            'crs': (rasterio.crs.CRS) crs of image proj
                            'raster_profile': (dict) geotiff profile of image
                            }
    :return: (tup) x,y pixel coords
    """
    geodata_copy = copy.deepcopy(geodata)
    crs = geodata_copy["crs"]

    # Convert point from EPSG:4326 to crs
    _transform = rasterio.Affine(*geodata_copy["transform"])
    # _transform = from_bounds(*bbox, width, height)
    pointx, pointy = rwarp.transform(
        rasterio.CRS.from_epsg(4326), crs, [point[1]], [point[0]]
    )

    # Get pixel coordinates
    x, y = rowcol(_transform, [pointx[0]], [pointy[0]])

    return int(y), int(x)


def get_crs(srctif):
    with rasterio.open(srctif, "r") as ds:
        crs = ds.crs
    return crs


def get_resolution(srctif):
    with rasterio.open(srctif, "r") as ds:
        _transform = ds.transform
        res_x = _transform[0]
        res_y = -_transform[4]

    return res_x, res_y


def transform_bbox(bbox, src_crs, dst_crs):
    """
    Transform bbox coordinates from source CRS to destination CRS

    @param bbox: (list) lat/lon image bounds [west, south, east, north]
    @param src_crs: (CRS object) Source CRS
    @param dst_crs: (CRS object) Destination CRS
    @return: (list) bbox coordinates in Destination CRS
    """

    x0, y0 = bbox[0], bbox[1]
    x1, y1 = bbox[2], bbox[3]

    x0t, y0t = rwarp.transform(src_crs, dst_crs, [x0], [y0])
    x1t, y1t = rwarp.transform(src_crs, dst_crs, [x1], [y1])

    bbox = [x0t[0], y0t[0], x1t[0], y1t[0]]

    return bbox


def get_max_bbox(base_bbox, overlaybbox):
    """
    Returns maximum common bounds of base and overlay sets
    @param: (list) bbox of base set
    @param: (list) bbox of overlay set
    @return: (list) maximum common bbox
    """
    _base = base_bbox
    base_box_adj = [[_base[0] + 180, _base[1] + 90, _base[2] + 180, _base[3] + 90]]
    overlay_box_adj = []
    for bbox in overlaybbox:
        overlay_box_adj = [[bbox[0] + 180, bbox[1] + 90, bbox[2] + 180, bbox[3] + 90]]

    bboxes = base_box_adj + overlay_box_adj
    bbox_polys = []
    for bbox in bboxes:
        bbox_polys.append(box(*bbox))

    _maxbbox = bbox_polys[0]
    for rect in bbox_polys:
        _maxbbox = _maxbbox.intersection(rect)

    maxbbox = _maxbbox.bounds
    maxbbox = [maxbbox[0] - 180, maxbbox[1] - 90, maxbbox[2] - 180, maxbbox[3] - 90]

    ret = BoundingBox(*maxbbox)
    return ret


def get_bbox_intersection(user_bbox, img_bbox):
    """
    Returns the intersection of bbox areas
    @param user_bbox: (list) user-provided bbox
    @param img_bbox: (list) image bbox
    @return: (list) bbox intersection
    """
    if user_bbox and img_bbox:
        user_bbox_adj = box(
            user_bbox[0] + 180, user_bbox[1] + 90, user_bbox[2] + 180, user_bbox[3] + 90
        )
        img_bbox_adj = box(
            img_bbox[0] + 180, img_bbox[1] + 90, img_bbox[2] + 180, img_bbox[3] + 90
        )
        intersection = img_bbox_adj.intersection(user_bbox_adj)

        if intersection.is_empty:
            ret = None
        else:
            _bounds = intersection.bounds
            ret = BoundingBox(
                _bounds[0] - 180, _bounds[1] - 90, _bounds[2] - 180, _bounds[3] - 90
            )
    else:
        ret = None

    return ret


def check_bbox(requested, actual, degtol=0, strict=False):
    """
    Check the user-requested bbox is valid. Assumes EPSG:4326 coordinates.
    @param requested: (list) user-requested bbox (west, south, east, north)
    @param actual: (list) image bbox (west, south, east, north)
    @param degtol: (float) allowed tolerance around coordinate in degrees
    @param strict: (bool) if true, check only that requested bbox is inside the actual bbox
    @return: (tup) tuple of bools
    """
    lon_ok = False
    lat_ok = False
    west_ok = False
    south_ok = False
    east_ok = False
    north_ok = False

    requested = [
        requested[0] + 180,
        requested[1] + 90,
        requested[2] + 180,
        requested[3] + 90,
    ]
    actual = [actual[0] + 180, actual[1] + 90, actual[2] + 180, actual[3] + 90]

    if not strict:
        if (actual[0] - degtol) <= requested[0] <= (actual[0] + degtol):
            west_ok = True
        if (actual[1] - degtol) <= requested[1] <= (actual[1] + degtol):
            south_ok = True
        if (actual[2] - degtol) <= requested[2] <= (actual[2] + degtol):
            east_ok = True
        if (actual[3] - degtol) <= requested[3] <= (requested[3] + degtol):
            north_ok = True
    else:
        if actual[0] <= requested[0]:
            west_ok = True
        if actual[1] <= requested[1]:
            south_ok = True
        if actual[2] >= requested[2]:
            east_ok = True
        if actual[3] >= requested[3]:
            north_ok = True

    if west_ok and east_ok:
        lon_ok = True
    if south_ok and north_ok:
        lat_ok = True

    return lat_ok, lon_ok


# ------ Image operations
def image_to_geotiff(img, geodata, outfile=None):
    """
    Converts a simple image (e.g. '*.png') to geotiff

    @param pngimg: (str) or (Path) path to png image to convert OR
                   (PIL.Image) source image object
    @param geodata: (dict) {'bbox': (tup) (west, south, east, north) in image crs,
                            'crs': (rasterio.crs.CRS) crs of image proj
                            'raster_profile': (dict) geotiff profile of image
                            }
    @param outfile: (str) save path for output image; if not provided a temporary file
                    will be created
    @return: (str) save path for output image
    """
    if isinstance(img, str):
        _path = img
    elif isinstance(img, Image.Image):
        _path = None
    else:
        _path = str(img)

    geodata_copy = copy.deepcopy(geodata)

    if _path:
        pngimg = Image.open(_path)
    else:
        pngimg = img

    # Convert image to array
    imgarr = np.array(pngimg)
    r = imgarr[:, :, 0]
    g = imgarr[:, :, 1]
    b = imgarr[:, :, 2]

    # Create new raster
    profile = geodata_copy["raster_profile"]
    width = profile["width"]
    height = profile["height"]
    dtype = profile["dtype"]
    if not geodata_copy["transform"]:
        _transform = from_bounds(*geodata_copy["bbox"], width, height)
    else:
        _transform = rasterio.Affine(*geodata_copy["transform"])

    if pngimg.mode == "RGB":
        count = 3
        imgbands = [r, g, b]
    else:
        count = 4
        a = imgarr[:, :, 3]
        imgbands = [r, g, b, a]

    kwargs = {}
    kwargs.update(
        {
            "transform": _transform,
            "width": width,
            "height": height,
            "crs": geodata_copy["crs"],
            "bounds": geodata_copy["bbox"],
            "count": count,
            "dtype": dtype,
        }
    )

    tmp_file = tempfile.NamedTemporaryFile(mode="w+b", delete=False, suffix=".tif")
    _file = tmp_file.name

    with rasterio.open(_file, "w", **kwargs) as dst:
        for band, src in enumerate(imgbands, start=1):
            dst.write_band(band, src)

    if outfile:
        os.replace(_file, outfile)
        return outfile
    else:
        return _file


def geotiff_to_image(srctif, outfile=None):
    """
    Convert a geotiff to a simple image
    @param srctif: (str) path to source tif
    @param outfile: (str) save path for output image; if not provided then image will
                    not be saved
    @return: (PIL.Image obj) output image
    """
    ds = rasterio.open(str(srctif), "r")
    bands = ds.indexes
    data = ds.read(bands)
    _transform = rasterio.transform.from_bounds(
        0, 0, 0, 0, data.shape[2], data.shape[1]
    )
    crs = get_crs(srctif)

    tmpfile = tempfile.NamedTemporaryFile("w+b", suffix=".png", delete=False)
    _file = tmpfile.name
    tmpfile.close()

    with rasterio.open(
        _file,
        "w",
        driver="PNG",
        width=data.shape[2],
        height=data.shape[1],
        count=len(bands),
        dtype=data.dtype,
        nodata=0,
        transform=_transform,
        crs=crs,
        format="PNG",
    ) as dst:
        dst.write(data, indexes=bands)

    img = Image.open(_file).convert("RGBA")
    if outfile:
        img.save(outfile)

    try:
        os.unlink(_file)
    except PermissionError:
        pass

    return img


def update_geodata(srctif, geodata):
    with rasterio.open(srctif, "r") as ds:
        profile = ds.profile
        crs = profile.pop("crs", None)
        _t = profile.pop("transform", None)
        geodata.update(
            {
                "raster_profile": profile.data,
                "crs": crs.wkt,
                "transform": [_t.a, _t.b, _t.c, _t.d, _t.e, _t.f],
                "bbox": ds.bounds,
                "res_x": ds.res[0],
                "res_y": ds.res[1],
            }
        )


def resize(srcpath, width, height, geodata=None, resample=None, outfile=None):
    """
    Resize geotiff
    @param srcpath: (str) or (Path) path to png or geotiff
    @param width: (int) desired output width
    @param height: (int) desired output height
    @param geodata: (dict) {'bbox': (tup) (west, south, east, north) in image crs,
                            'crs': (rasterio.crs.CRS) crs of image proj
                            'raster_profile': (dict) geotiff profile of image
                            }
    @param resample: (rasterio.enums.Resampling) default is Resampling.nearest
    @param delete: (bool) delete input file and replace with resized output file
    @param outfile: (str) save path for output image; if not provided and delete is False
                    then a temporary file will be created
    @return: (tup) ((str) save path for output image, (dict) updated geodata [for png only])
    """
    # Accept Path obj for srcpath argument
    if isinstance(srcpath, str):
        _path = srcpath
    else:
        _path = str(srcpath)

    # Determine suffix of srcpath
    if _path.split(".")[-1] == "png":
        is_gtiff = False
        converted = False
        geodata_copy = copy.deepcopy(geodata)
        profile_out = geodata_copy["raster_profile"]
    else:
        is_gtiff = True
        converted = False

    if not resample:
        resample = Resampling.nearest

    # Create tempfiles for writing
    tmpfile = tempfile.NamedTemporaryFile("w+b", suffix=".tif", delete=False)
    _file = tmpfile.name
    tmpfile.close()

    tmpfile2 = tempfile.NamedTemporaryFile("w+b", suffix=".tif", delete=False)
    _file2 = tmpfile2.name
    tmpfile2.close()

    while True:
        if is_gtiff or converted:
            with rasterio.open(_path, "r") as ds:
                data = ds.read(
                    out_shape=(ds.count, int(height), int(width)), resampling=resample
                )
                _transform = ds.transform * ds.transform.scale(
                    (ds.width / data.shape[-1]), (ds.height / data.shape[-2])
                )
                profile = ds.profile

                if not is_gtiff:
                    profile = profile | profile_out
                    profile_out.update({"height": height, "width": width})

                profile.update(
                    {
                        "count": ds.count,
                        "height": height,
                        "width": width,
                        "transform": _transform,
                    }
                )

                with rasterio.open(_file2, "w", **profile) as dst:
                    dst.write(data)

            if not is_gtiff:
                res_x, res_y = get_resolution(_file2)
                geodata_copy.update({"res_x": res_x, "res_y": res_y})
                geodata_copy.update(
                    {
                        "transform": [
                            _transform.a,
                            _transform.b,
                            _transform.c,
                            _transform.d,
                            _transform.e,
                            _transform.f,
                        ]
                    }
                )
            try:
                os.unlink(_file)
            except PermissionError:
                pass
            break
        else:
            _path = image_to_geotiff(_path, geodata_copy)
            converted = True

    if not is_gtiff:
        if outfile:
            geotiff_to_image(_file2, outfile)
            try:
                os.unlink(_file2)
                os.unlink(_path)
            except PermissionError:
                pass
            retpath = outfile
        else:
            outfilepath = tempfile.NamedTemporaryFile(
                "w+b", suffix=".png", delete=False
            )
            outfilepath.close()
            retpath = outfilepath.name
            geotiff_to_image(_file2, retpath)
            try:
                os.unlink(_file2)
                os.unlink(_path)
            except PermissionError:
                pass
    else:
        if outfile:
            os.replace(_file2, outfile)
            retpath = outfile
        else:
            retpath = _file2

    if not is_gtiff:
        return retpath, geodata_copy
    else:
        return retpath, None


def get_shape(srcpath, geodata=None):
    """
    Get the shape in pixels of a geotiff
    @param srcpath: (str) or (Path) path to png or geotiff image file
    @return: (tup) width in pixels, height in pixels
    """
    if isinstance(srcpath, str):
        _path = srcpath
    else:
        _path = str(srcpath)

    if srcpath.split(".")[-1] == "png":
        is_gtiff = False
        converted = False
        geodata_copy = copy.deepcopy(geodata)
    else:
        is_gtiff = True
        converted = False

    while True:
        if is_gtiff and converted:
            with rasterio.open(_path) as src:
                w = src.shape[1]
                h = src.shape[0]
            break
        else:
            _path = image_to_geotiff(_path, geodata_copy)
            converted = True

    return w, h


def reproject(srcpath, out_crs="EPSG:4326", geodata=None, outfile=None):
    """
    Reproject a geotiff or png to specified CRS.
    @param srcpath: (str) or (Path) path to png or geotiff image
    @param out_crs: (str) or (rasterio.CRS obj) desired output CRS of image
    @param geodata: (dict) {'bbox': (tup) (west, south, east, north) in image crs,
                            'crs': (rasterio.crs.CRS) crs of image proj
                            'raster_profile': (dict) geotiff profile of image
                            }
    @param outfile: (str) output .tif file name, if not given then a temporary file
                          will be created
    @return: (tup) ((str) save path for output image, (dict) updated geodata [for png only])
    """
    if isinstance(srcpath, str):
        _path = srcpath
    else:
        _path = str(srcpath)

    if _path.split(".")[-1] == "png":
        is_gtiff = False
        converted = False
        geodata_copy = copy.deepcopy(geodata)
        profile_out = geodata_copy["raster_profile"]
    else:
        is_gtiff = True
        converted = False

    tmpgtifffile = tempfile.NamedTemporaryFile("w+b", suffix=".tif", delete=False)
    gtiff_file = tmpgtifffile.name

    while True:
        if is_gtiff or converted:
            with rasterio.open(_path) as src:
                if isinstance(out_crs, str):
                    out_crs = rasterio.CRS.from_string(out_crs)
                else:
                    out_crs = out_crs

                _transform, width, height = rwarp.calculate_default_transform(
                    src.crs, out_crs, src.width, src.height, *src.bounds
                )

                if not is_gtiff:
                    bbox_out = transform_bbox(
                        geodata_copy["bbox"], geodata_copy["crs"], out_crs
                    )
                    # old_width = geodata_copy['raster_profile']['width']
                    # old_height = geodata_copy['raster_profile']['height']
                    # if old_width != width or old_height != height:
                    #    src2_path, _ = resize(_path, old_width, old_height, geodata=geodata_copy)
                    # else:
                    #    src2_path = None
                else:
                    bbox_out = transform_bbox(src.bounds, src.crs, out_crs)
                    # if width != src.width or height != src.height:
                    #    src2_path, _ = resize(_path, src.width, src.height)
                    # else:
                    #    src2_path = None

                profile = src.profile

                if not is_gtiff:
                    geodata_copy.update(
                        {
                            "crs": out_crs,
                            "bbox": bbox_out,
                            "transform": [
                                _transform.a,
                                _transform.b,
                                _transform.c,
                                _transform.d,
                                _transform.e,
                                _transform.f,
                            ],
                        }
                    )
                    profile_out.update({"width": width, "height": height})

                profile.update({"transform": _transform})
                profile.update(
                    {
                        "crs": out_crs,
                        "bounds": bbox_out,
                        "count": src.count,
                        "width": width,
                        "height": height,
                    }
                )

                src2_path = gtiff_file

                with rasterio.open(src2_path, "w", **profile) as dst:
                    for i in range(1, src.count + 1):
                        rwarp.reproject(
                            rasterio.band(src, i),
                            destination=rasterio.band(dst, i),
                            src_transform=src.transform,
                            src_crs=src.crs,
                            dst_transform=_transform,
                            dst_crs=out_crs,
                            resampling=Resampling.nearest,
                        )

                if not is_gtiff:
                    res_x, res_y = get_resolution(src2_path)
                    geodata_copy.update({"res_x": res_x, "res_y": res_y})
            break
        else:
            _path = image_to_geotiff(_path, geodata_copy)
            converted = True

    tmpgtifffile.close()
    if not is_gtiff:
        if outfile:
            geotiff_to_image(src2_path, outfile=outfile)
            try:
                os.unlink(src2_path)
            except PermissionError:
                pass
            retpath = outfile
        else:
            img = geotiff_to_image(src2_path)
            tmpimgfile = tempfile.NamedTemporaryFile("w+b", suffix=".png", delete=False)
            imgfile = tmpimgfile.name
            img.save(imgfile)
            img.close()
            try:
                os.unlink(src2_path)
            except PermissionError:
                pass
            retpath = imgfile
    else:
        if outfile:
            os.replace(src2_path, outfile)
            retpath = outfile
        else:
            retpath = src2_path

    if converted:
        try:
            os.unlink(_path)
        except PermissionError:
            pass

    if not is_gtiff:
        return retpath, geodata_copy
    else:
        return retpath, None


def get_bounds(srcpath):
    """
    Get the bounds of a geotiff
    @param srcpath: (str) or (Path) path to geotiff
    @return: (list) bbox bounds
    """
    if isinstance(srcpath, str):
        _path = srcpath
    else:
        _path = str(srcpath)

    with rasterio.open(_path) as src:
        bounds = src.bounds

    return bounds


def crop(srcpath, new_bbox, geodata=None, outfile=None):
    """
    Crop a geotiff to specified bbox and also resize, if desired.
    @param srcpath: (str) or (Path) path to png or geotiff image
    @param new_bbox: (list) bbox bounds
    @param geodata: (dict) {'bbox': (tup) (west, south, east, north) in image crs,
                            'crs': (rasterio.crs.CRS) crs of image proj
                            'raster_profile': (dict) geotiff profile of image
                            }
    @param outfile: (str) output .tif file name, if not given then a temporary file
                          will be created
    @return: (tup) ((str) save path for output image, (dict) updated geodata [for png only])
    """
    if isinstance(srcpath, str):
        _path = srcpath
    else:
        _path = str(srcpath)

    if _path.split(".")[-1] == "png":
        is_gtiff = False
        converted = False
        geodata_copy = copy.deepcopy(geodata)
        profile_out = geodata_copy["raster_profile"]
    else:
        is_gtiff = True
        converted = False

    tmpfile = tempfile.NamedTemporaryFile("w+b", suffix=".tif", delete=False)
    gtiff_file = tmpfile.name

    # Need to transform bbox if crs is not EPSG:4326
    if not is_gtiff:
        if geodata_copy["crs"] != "EPSG:4326":
            new_bbox = transform_bbox(new_bbox, "EPSG:4326", geodata_copy["crs"])
    else:
        with rasterio.open(_path, "r") as ds:
            crs = ds.crs
        if crs != "EPSG:4326":
            new_bbox = transform_bbox(new_bbox, "EPSG:4326", crs)

    while True:
        if is_gtiff or converted:
            with rasterio.open(_path, "r") as src:
                bands = len(src.indexes)
                cropped = src.read(
                    bands, window=rwindows.from_bounds(*new_bbox, src.transform)
                )

                profile = src.profile

                if not is_gtiff:
                    profile = profile | profile_out
                    profile_out.update(
                        {"height": cropped.shape[0], "width": cropped.shape[1]}
                    )

                profile.update(
                    {
                        "bounds": new_bbox,
                        "height": cropped.shape[0],
                        "width": cropped.shape[1],
                    }
                )

                with rasterio.open(gtiff_file, "w", **profile) as dst:
                    for band in range(1, bands + 1):
                        dst.write_band(
                            band,
                            src.read(
                                band,
                                window=rwindows.from_bounds(*new_bbox, src.transform),
                            ),
                        )

            if not is_gtiff:
                res_x, res_y = get_resolution(gtiff_file)
                geodata_copy.update({"bbox": new_bbox, "res_x": res_x, "res_y": res_y})
            break
        else:
            _path = image_to_geotiff(_path, geodata_copy)
            converted = True

    tmpfile.close()
    if not is_gtiff:
        if outfile:
            geotiff_to_image(gtiff_file, outfile=outfile)
            try:
                os.unlink(gtiff_file)
            except PermissionError:
                pass
            retpath = outfile
        else:
            img = geotiff_to_image(gtiff_file)
            tmpimgfile = tempfile.NamedTemporaryFile("w+b", suffix=".png", delete=False)
            imgfile = tmpimgfile.name
            img.save(imgfile)
            img.close()
            try:
                os.unlink(gtiff_file)
            except PermissionError:
                pass
            retpath = imgfile
    else:
        if outfile:
            os.replace(gtiff_file, outfile)
            retpath = outfile
        else:
            retpath = gtiff_file

    if converted:
        try:
            os.unlink(_path)
        except PermissionError:
            pass

    if not is_gtiff:
        return retpath, geodata_copy
    else:
        return retpath, None


def adjust_alpha(srcpath, alpha, outfile=None):
    """
    Adjust alpha of simple image
    @param srcpath: (str) or (Path) path to png or geotiff input image
    @param alpha: (float) desired alpha value
    @param outfile: (str)
    @return: (str) path to modified file in png format
    """
    if isinstance(srcpath, str):
        _path = srcpath
    else:
        _path = str(srcpath)

    if _path.split(".")[-1] == "png":
        is_gtiff = False
        converted = False
        img = Image.open(_path)
    else:
        is_gtiff = True
        converted = False

    tmpfile = tempfile.NamedTemporaryFile("w+b", suffix=".png", delete=False)
    png_file = tmpfile.name

    while True:
        if not is_gtiff or converted:
            imgarr = np.array(img, dtype=np.uint8)
            np.place(imgarr[:, :, 3], imgarr[:, :, 3] > 0, int(alpha * 255))
            retimg = Image.fromarray(imgarr, "RGBA")
            retimg.save(png_file)
            retimg.close()
            break
        else:
            img = geotiff_to_image(_path)
            converted = True

    tmpfile.close()
    if outfile:
        os.replace(png_file, outfile)
        try:
            os.unlink(png_file)
        except PermissionError:
            pass
        retpath = outfile
    else:
        retpath = png_file

    return retpath


def trim_img_border(img):
    """
    Trim border of simple image
    @param img: (PIL.Image obj) input image
    @return: (PIL.Image obj) trimmed output image
    """
    imgarr = np.array(img, dtype=np.uint8)
    mask = np.array((255 - imgarr[:, :, 3]) / 255, dtype=np.uint8)
    r = np.ma.array(imgarr[:, :, 0], mask=mask)
    b = np.ma.array(imgarr[:, :, 1], mask=mask)
    g = np.ma.array(imgarr[:, :, 2], mask=mask)

    unmasked_slices = np.ma.notmasked_contiguous(r, axis=1)
    top_row_count = 0
    bot_row_count = 0
    for _slice in unmasked_slices:
        if not _slice:
            top_row_count += 1
        else:
            break

    for _slice in unmasked_slices[::-1]:
        if not _slice:
            bot_row_count += 1
        else:
            break

    top_row = top_row_count
    bot_row = imgarr.shape[0] - bot_row_count

    start_cols = []
    stop_cols = []
    for _slice in unmasked_slices:
        if _slice:
            start_cols.append(_slice[0].start)
            stop_cols.append(_slice[0].stop)

    left_col = min(start_cols)
    right_col = max(stop_cols)

    r = r[top_row:bot_row, left_col:right_col]
    b = b[top_row:bot_row, left_col:right_col]
    g = g[top_row:bot_row, left_col:right_col]

    newarr = np.dstack((r, b, g))

    retimg = Image.fromarray(newarr, "RGB")
    return retimg


# ------ Misc functions
def get_timestamp(file, tformatdict, tz=pytz.utc):
    """
    Applies user-provided time format string to filename to extract timestamp.
    The returned datetime also has its timezone set to that provided, or UTC if
    none is provided
    @param file: (str) filename containing datetime string
    @param tformatdict: (dict) indexes returned by _time_format_str
    @param tz: (pytz tz obj) pytz timezone object
    @return: (datetime) tz-aware datetime
    """

    kwargs = {}
    if tformatdict["Y"][0] is not None and tformatdict["Y"][1] is not None:
        year = int(file[tformatdict["Y"][0] : tformatdict["Y"][1]])
    else:
        year = None
    if tformatdict["M"][0] is not None and tformatdict["M"][1] is not None:
        month = int(file[tformatdict["M"][0] : tformatdict["M"][1]])
    else:
        month = None
    if tformatdict["D"][0] is not None and tformatdict["D"][1] is not None:
        day = int(file[tformatdict["D"][0] : tformatdict["D"][1]])
    else:
        day = None
    if tformatdict["h"][0] and tformatdict["h"][1]:
        kwargs["hour"] = int(file[tformatdict["h"][0] : tformatdict["h"][1]])
    if tformatdict["m"][0] and tformatdict["m"][1]:
        kwargs["minute"] = int(file[tformatdict["m"][0] : tformatdict["m"][1]])
    if tformatdict["s"][0] and tformatdict["s"][1]:
        kwargs["second"] = int(file[tformatdict["s"][0] : tformatdict["s"][1]])

    if not year or not month or not day:
        print(
            f"{Fore.RED}WARNING - Could not extract time from provided filename {file}. Check time "
            f"format string."
        )
        t = None

    t = tz.localize(datetime(year, month, day, **kwargs))

    return t


def time_format_str(instr):
    """
    Generates time format dictionary to be used for parsing image filenames
    @param instr: (str) input format string (e.g. "YYYY-MM-dd hh_mm_ss")
    @return: (dict) time format dict
    """
    if instr:
        chars = ["Y", "M", "D", "h", "m", "s"]
        start_idxs = []
        end_idxs = []
        start_done = False
        char_found = False

        for c in chars:
            for i, _s in enumerate(instr):
                if _s == c and not start_done:
                    start_idxs.append(i)
                    start_done = True
                    char_found = True
                elif _s != c and start_done:
                    end_idxs.append(i)
                    start_done = False
                elif _s == c and start_done and i == len(instr) - 1:
                    end_idxs.append(i)
            if not char_found:
                start_idxs.append(None)
                end_idxs.append(None)
            else:
                char_found = False

        retdict = {}
        for i, c in enumerate(chars):
            retdict[c] = (start_idxs[i], end_idxs[i])
    else:
        retdict = None

    return retdict


def replace_file_timezone(file, in_tz, out_tz, out_abbr):
    """
    Renames a file to match desired timezone
    @param file: (str) path to file
    @param in_tz: (pytz tz obj) timezone in input filename
    @param out_tz: (pytz tz obj) timezone in output filename
    @param out_abbr: (str) timezone abbreviation to use in output filename
    @return: None
    """
    if isinstance(file, str):
        tstr = file.split("\\")[-1].split(".")[0]
        svpath = "\\".join(file.split("\\")[0:-1])
        svpath = svpath.rstrip("\\") + "\\"
    elif isinstance(file, Path):
        tstr = file.stem
        svpath = file.parent
    else:
        print(
            f"{Fore.RED}WARNING - File path must be provided as a string or Path object"
        )
        return

    if not isinstance(in_tz, pytz.BaseTzInfo):
        print(
            f"{Fore.RED}WARNING - Timezone must be provided as a pytz timezone object."
        )
        return
    if not isinstance(out_tz, pytz.BaseTzInfo):
        print(
            f"{Fore.RED}WARNING - Timezone must be provided as a pytz timezone object."
        )
        return

    _s1 = tstr.split(" ")[0]
    _s2 = tstr.split(" ")[1]
    year = int(_s1.split("-")[0])
    month = int(_s1.split("-")[1])
    day = int(_s1.split("-")[2])
    hr = int(_s2.split("_")[0])
    _min = int(_s2.split("_")[1])
    sec = int(_s2.split("_")[2])

    t = datetime(year, month, day, hour=hr, minute=_min, second=sec)
    t2 = in_tz.localize(t)
    t3 = t2.astimezone(out_tz)
    newtstr = t3.isoformat()
    newtstr = newtstr.replace("T", " ").replace(":", "_")
    newtstr = newtstr[0:-6] + " " + out_abbr

    if isinstance(file, str):
        fname = svpath + newtstr
    else:
        fname = svpath / newtstr

    os.replace(file, fname)
    return



import atexit
import copy
import os
import shutil
import sys
import tempfile
from datetime import datetime
from pathlib import Path

import numpy as np
import pytz
from PIL import Image
from colorama import Fore
from matplotlib import colormaps as cm
from moviepy.editor import ImageSequenceClip
from tqdm import tqdm

from goesvideo.utils import gistools, editortools

"""
The GenericOverlay class is intended to produce composite images from a base set of images (typically a 
static/dynamic visible satellite image of an area or a map) and one or more overlay sets 
(e.g. fire detection). The image sets should all cover the same time frame, but they do not require 
exactly matching timestamps. The class will set the time interval between frames to the lowest 
possible value given base and overlay image sets with each set assumed to contain images with a 
constant time interval. The image sets with higher time intervals will be filled with redundant 
frames as required to match the lowest interval.
"""


class GenericOverlay:
    def __init__(
            self,
            baseimgpath,
            overlaypaths,
            overlay_timezones=None,
            base_timezone=None,
            start_time=None,
            end_time=None,
    ):
        """
        @param baseimgpath: (str) path to folder or file to use for base image(s). If a folder is
                                  provided the object instance will be initialized using all
                                  available geotiffs or pngs in the folder and the base image will
                                  be dyanmic. If a single file or a folder containing a single file
                                  is provided, the base image will be static
        @param overlaypaths: (list) paths to folders containing overlay images. Provide one path for
                                    each desired overlay set; *images must be geotiff format*
        @param base_timezone: (dict) {'timezone': (pytz tz obj), 'filename format': (str)} Image
                                     filenames must be provided as datetimes. The format can be
                                     anything as long as it is specified here
                                     (e.g. "YYYY-MM-DD HH:mm:ss"). If parameter is not specified,
                                     UTC will be used for the timezone and the function will attempt
                                     to convert the filenames on its own assuming isoformat strings
                                     with dash and underscore separators ("YYYY-MM-DD hh_mm_ss").
        @param overlay_timezones: (dict) {'timezone': [(pytz tz obj)], 'filename format': [(str)]}
                                         Image filenames must be provided as datetimes. The format
                                         can be anything as long as it is specified here
                                         (e.g. "YYYY-MM-DD HH:mm:ss"). If parameter is not
                                         specified, UTC will be used for the timezone and the
                                         function will attempt to convert the filenames assuming
                                         isoformat strings. Provide one timezone and one format for
                                         each overlay set.
        @param start_time: (str) isoformat datetime start time, by default, timezone is assumed to
                                 be the same as that provided by base_timezone. This behavior can
                                 be overridden by including the offset in the datetime string
                                 (e.g. '2023-01-01 12:00:00-03:00'). If not provided, the default
                                 behavior is to use all available images in specified folders.
        @param end_time: (str) isoformat datetime end time, by default, timezone is assumed to be
                               the same as that provided by base_timezone. This behavior can be
                               overridden by including the offset in the datetime string
                               (e.g. '2023-01-01 12:00:00-03:00'). If not provided, the default
                               behavior is to use all available images in specified folders.

        """
        # Check start and end time parameters
        if not start_time and not end_time:
            start_time = "1900-01-01 00:00:00"
            end_time = "2100-01-01 00:00:00"
        elif not start_time:
            start_time = "1900-01-01 00:00:00"
        elif not end_time:
            end_time = "2100-01-01 00:00:00"

        # Setup temporary working folders for base and overlay images
        print(f"{Fore.GREEN}Initalizing working folders...", end="")
        self._setup_working_folders(overlaypaths)
        atexit.register(self.cleanup)
        print(f"{Fore.GREEN}Done!")

        print(
            f"{Fore.GREEN} Filtering images and copying to working folders..."
        )
        # Copy base image files to working folder and filter list of filenames to match start and
        # end times
        self._setup_base_imgs(baseimgpath, base_timezone, start_time, end_time)

        # Copy overlay image files to working folders and filter filename lists to match start and
        # end times
        self._setup_overlay_imgs(
            overlaypaths, overlay_timezones, start_time, end_time)

        # Match lengths of overlay image arrays to the length of the base array
        self._setup_matched_lists()
        print(f"{Fore.GREEN} Done!")

        # Get projections and bounds for base and overlays
        if self.base_is_geotiff:
            self.base_crs = gistools.get_crs(self.baseimgfiles[0])
            self.base_bbox = gistools.get_bounds(self.baseimgfiles[0])
            self.base_bbox_latlon = gistools.transform_bbox(
                self.base_bbox, self.base_crs, "EPSG:4326"
            )
        else:
            self.base_crs = None

        self.overlay_crs = []
        self.overlay_bbox = []
        self.overlay_bbox_latlon = []
        for i, overlay in enumerate(self.overlayimgfiles):
            self.overlay_crs.append(gistools.get_crs(overlay[0]))
            self.overlay_bbox.append(gistools.get_bounds(overlay[0]))
            self.overlay_bbox_latlon.append(
                gistools.transform_bbox(
                    self.overlay_bbox[i], self.overlay_crs[i], "EPSG:4326"
                )
            )
        if not self.base_is_geotiff:
            self.base_crs = self.overlay_crs[0]
            self.base_bbox_latlon = self.overlay_bbox_latlon[0]

        print()
        print(f"{Fore.GREEN} Initialization Complete!")
        print()

    def create_overlays(
            self,
            outpath,
            out_crs=None,
            res=None,
            bbox=None,
            overlay_opacities=None,
            overlay_resample=False,
            base_resample=False,
            save_overlays_gtiff=None,
            cumulative=False,
            cumulative_colormap=None,
            output_format="simple_image",
    ):
        """

        @param outpath: (str) path to save output images
        @param out_crs: (rasterio.crs.CRS obj) CRS to use for output images
        @param res: (tup) desired output resolution (width, height)
        @param bbox: (list) [minx, maxy, maxx, miny] in lat/lon coords. If provided, an attempt will
                            be made to crop the base and overlay images to specified bounds. In the
                            event that the provided bbox overlaps partially or not at all with one
                            or more image sets, the default behavior will be to find the maximum
                            common bounds shared by all the image sets. The maximum common bounds
                            will also be used if the bbox parameter is not specified. Note that if
                            pngs are used for the base image set, the assumption is that the user
                            has ensured that the image extents already match those
                            defined by the overlay geotiffs.
        @param overlay_opacities: (list) Opacities of overlay layers. Provide one value (0-1) for
                                         each overlay set. Default is value of 1.
        @param overlay_resample: (bool) Enable resampling of overlay image set(s) upon resizing
        @param base_resample:  (bool) Enable resampling of base image set upon resizing
        @param save_overlays_gtiff: (str) Path to folder for saving individual overlay frames as
                                          separate geotiffs. Default is to discard after compositing
                                          with base image.
        @param cumulative: (bool) If true, a cumulative overlay sequence will be generated over the
                                  base frame(s)
        @param cumulative_colormap: (matplotlib.cm obj) If a colormap is provided, the frame
                                    sequence will automatically be set to cumulative and the color
                                    applied to each overlay will be incremented at each frame.
        @param output_format: (str) options are 'simple_image' which will produce pngs or 'geotiff'
        @return: None
        """
        outpath = Path(outpath)

        if not overlay_opacities:
            overlay_opacities = [1] * len(self.overlaypaths)

        if base_resample:
            base_kwargs = {"resample": True}
        else:
            base_kwargs = {}

        if overlay_resample:
            overlay_kwargs = {"resample": True}
        else:
            overlay_kwargs = {}

        # If no output resolution is specified, use that of the base image
        if not res:
            if self.base_is_geotiff:
                res = gistools.get_shape(str(self.baseimgfiles[0]))
            else:
                res = Image.open(str(self.baseimgfiles[0])).size

        # Check user-provided bbox to make sure it at least overlaps partially with the base/overlay
        # images. If not, default to using the maximal common bounds.
        # If bbox parameter is not provided, default to maximal common bounds
        if bbox and self.base_is_geotiff:
            _base_bbox = gistools.get_bbox_intersection(
                bbox, self.base_bbox_latlon)
            _overlay_bbox = []

            for overlaybbox in self.overlay_bbox_latlon:
                _overlay_bbox.append(
                    gistools.get_bbox_intersection(bbox, overlaybbox))

            _bbox_list = [_base_bbox] + _overlay_bbox
            err1 = (
                f"{Fore.RED} WARNING! Provided bbox is outside the bounds of the base and/or "
                f"overlay images. Using the maximal common bounds of the images instead."
            )
            err2 = (
                f"{Fore.RED} WARNING! Provided bbox only overlaps partially with the base and/or"
                f" overlay images. Final images will be smaller than expected."
            )
            err3 = (
                f"{Fore.RED} WARNING! Provided bbox does not overlap with all overlay sets. Using "
                f"maximal common bounds of the images instead."
            )

            if not all([x == bbox for x in _bbox_list]):
                if any([x is None for x in _bbox_list]):
                    print(err1)
                    bbox = gistools.get_max_bbox(
                        self.base_bbox_latlon, self.overlay_bbox_latlon
                    )
                else:
                    print(err2)
                    bbox1 = gistools.get_bbox_intersection(bbox, _base_bbox)
                    for i, ovbbox in enumerate(_overlay_bbox):
                        bbox1 = gistools.get_bbox_intersection(bbox1, ovbbox)
                    if not bbox1:
                        print(err3)
                        bbox = gistools.get_max_bbox(
                            self.base_bbox_latlon, self.overlay_bbox_latlon
                        )
                    else:
                        bbox = bbox1
        elif bbox and not self.base_is_geotiff:
            print(
                f"{Fore.RED} WARNING! Simple image(s) provided as base. Overlay images are assumed "
                f"to match the lat/lon bounds of the base image(s). For most accurate results, "
                f"provide geotiff(s) for base set."
            )
        else:
            bbox = gistools.get_max_bbox(
                self.base_bbox_latlon, self.overlay_bbox_latlon
            )

        # If user does not provide an output CRS then try to use the base image CRS first; if that
        # isn't available (e.g. due to pngs being used for the base image) then default to the CRS
        # of the first overlay set
        if not out_crs:
            if self.base_crs:
                out_crs = self.base_crs
            else:
                out_crs = self.overlay_crs[0]

        # Reproject base and overlay images if necessary to match the specified output CRS
        if self.base_is_geotiff:
            if self.base_crs != out_crs:
                print(
                    f"{Fore.RED} WARNING! Projection of base image(s) does not match the output "
                    f"CRS. Converting..."
                )
                for f in tqdm(self.baseimgfiles, desc="Image Number: ", colour="blue"):
                    gistools.reproject(f, out_crs)
                print(f"{Fore.RED} Done!")

        for i, overlay in enumerate(self.overlayimgfiles):
            if self.overlay_crs[i] != out_crs:
                print(
                    f"{Fore.RED} WARNING! Projection of overlay images at index {str(i)} does not "
                    f"match the output CRS. Converting..."
                )
                for f in tqdm(overlay, desc="Image Number: ", colour="blue"):
                    gistools.reproject(f, out_crs)
                print(f"{Fore.RED} Done!")

        # Transform bbox coordinates if necessary to match output CRS
        if out_crs != "EPSG:4326":
            outbbox = gistools.transform_bbox(bbox, "EPSG:4326", out_crs)
        else:
            outbbox = bbox

        # Crop base and overlay images to bbox coordinates and resize
        imgheight = res[1]
        imgwidth = res[0]
        if self.base_is_geotiff:
            print(f"{Fore.GREEN} Cropping base and overlay images...", end="")
            if self.base_is_geotiff:
                for f in self.baseimgfiles:
                    gistools.crop_geotiff(
                        f, outbbox, imgheight, imgwidth, opacity=1, **base_kwargs
                    )
            else:
                for f in self.baseimgfiles:
                    img = Image.open(f)
                    img.resize((imgwidth, imgheight))
                    img.save(f)
                    img.close()

            for i, overlay in enumerate(self.overlayimgfiles):
                for f in overlay:
                    gistools.crop_geotiff(
                        f,
                        outbbox,
                        imgheight,
                        imgwidth,
                        opacity=overlay_opacities[i],
                        **overlay_kwargs,
                    )

        print(f"{Fore.GREEN} Done!")

        # Final setup stuff
        if cumulative_colormap:
            cmap = cm.get_cmap(cumulative_colormap)
        else:
            cmap = None

        if save_overlays_gtiff:
            savepath = Path(save_overlays_gtiff)
            savepath.mkdir(exist_ok=True)
            overlaypaths = []
            for i in range(len(self.overlayimgfiles)):
                op = savepath / f"overlay_{str(i)}"
                op.mkdir(exist_ok=True)
                overlaypaths.append(op)

        # Image generation loop
        basedone = False
        img_arr_last = []
        print(f"{Fore.GREEN} Preparing overlay frames...")
        for i in tqdm(range(0, self.framecount), desc="Frame Number: ", colour="green"):
            img_set = []
            img_arr = []

            for row in self.overlayimgfiles:
                img_set.append(row[i])

            # Get base image path
            if self.base_is_static:
                baseimgpath = self.baseimgfiles[0]
            else:
                baseimgpath = self.baseimgfiles[i]

            # Generate save name for image
            if self.base_is_reference:
                if output_format == "simple_image":
                    svname = baseimgpath.stem + ".png"
                elif output_format == "geotiff":
                    svname = baseimgpath.stem + ".tif"
            else:
                if output_format == "simple_image":
                    svname = (
                        self.overlayimgfiles[self.overlay_reference_idx][i].stem
                        + ".png"
                    )
                elif output_format == "geotiff":
                    svname = (
                        self.overlayimgfiles[self.overlay_reference_idx][i].stem
                        + ".tif"
                    )

            # Convert base geotiff to image or just open image if it is a png
            if self.base_is_geotiff:
                if not basedone:
                    base_img = gistools.geotiff_to_image(baseimgpath)
                    if self.base_is_static:
                        basedone = True
            else:
                if not basedone:
                    base_img = Image.open(baseimgpath).convert("RGBA")
                    if self.base_is_static:
                        basedone = True

            if not img_arr_last:
                img_arr_last = [
                    Image.new(mode="RGBA", size=(imgwidth, imgheight)) for x in img_set
                ]

            for overlay, file in enumerate(img_set):
                if save_overlays_gtiff:
                    fname = overlaypaths[overlay] / \
                        (svname.split(".")[0] + ".tif")
                    shutil.copy(file, fname)

                # Convert overlay geotiffs to images and resize
                overlayimg = gistools.geotiff_to_image(file)
                if not self.base_is_geotiff:
                    overlayimg = overlayimg.resize((imgwidth, imgheight))

                # Apply colormap to overlay
                if cumulative_colormap:
                    pixels = overlayimg.load()
                    fname = file.stem
                    tstamp = gistools.get_timestamp(
                        fname, self.overlay_tformats[overlay], self.overlays_tz[overlay]
                    )
                    tnorm = (
                        tstamp - self.overlay_start[overlay]
                    ).total_seconds() / self.overlay_deltas[overlay]
                    color = cmap(tnorm)

                    for j in range(overlayimg.size[0]):
                        for k in range(overlayimg.size[1]):
                            if pixels[j, k][3] != 0:
                                pixels[j, k] = (
                                    int(color[0] * 255),
                                    int(color[1] * 255),
                                    int(color[2] * 255),
                                    pixels[j, k][3],
                                )

                img_arr.append(overlayimg)

            base_img_copy = copy.deepcopy(base_img)
            if cumulative:
                cum_arr = []
                for _img1, _img2 in zip(img_arr, img_arr_last):
                    _img1.alpha_composite(_img2)
                    cum_arr.append(_img1)
                img_arr = cum_arr

            for _img in img_arr:
                base_img_copy.alpha_composite(_img)

                if not outpath.exists():
                    outpath.mkdir()

                if output_format == "simple_image":
                    base_img_copy.save(str(outpath / svname))
                elif output_format == "geotiff":
                    fname = gistools.image_to_geotiff(
                        base_img_copy, bbox, out_crs)
                    shutil.copy(fname, str(outpath / svname))
                    os.remove(fname)

            img_arr_last = copy.deepcopy(img_arr)

        print(f"{Fore.GREEN} Overlays Complete!")

        return

    def cleanup(self):
        shutil.rmtree(str(self.basepath))
        for p in self.overlaypaths:
            shutil.rmtree(str(p))

    def _setup_working_folders(self, overlaypaths):
        self.basetmpdir = tempfile.TemporaryDirectory()
        self.basepath = Path(self.basetmpdir.name)
        self.overlaytmpdirs = []
        self.overlaypaths = []
        for p in overlaypaths:
            _path = tempfile.TemporaryDirectory()
            self.overlaytmpdirs.append(_path)
            self.overlaypaths.append(Path(_path.name))

    def _setup_base_imgs(self, baseimgpath, base_timezone, start_time, end_time):
        self.baseimgfiles = []
        baseimgpath = Path(baseimgpath)

        if baseimgpath.is_dir():
            gtiffs = list(baseimgpath.glob("*.tif")) + \
                list(baseimgpath.glob("*.tiff"))
            gtiffs = sorted(gtiffs, key=lambda item: item.name)
            pngs = list(baseimgpath.glob("*.png"))
            pngs = sorted(pngs, key=lambda item: item.name)
        else:
            if baseimgpath.suffix == ".tif" or baseimgpath.suffix == ".tiff":
                gtiffs = [baseimgpath]
                pngs = []
            elif baseimgpath.suffix == ".png":
                gtiffs = []
                pngs = [baseimgpath]
            else:
                print(f"{Fore.RED} ERROR: No valid image formats found. Exiting...")
                sys.exit(0)

        # Configure timezones and filename formats for base image(s) and overlay images
        if base_timezone:
            self.base_tz = base_timezone.pop("timezone", pytz.utc)
            base_format = base_timezone.pop("filename format", None)
            if base_format:
                self.base_tformat = gistools.time_format_str(base_format)
            else:
                self.base_tformat = gistools.time_format_str(
                    "YYYY-MM-DD hh_mm_ss")
        else:
            self.base_tz = pytz.utc
            self.base_tformat = gistools.time_format_str("YYYY-MM-DD hh_mm_ss")

        # Filter base image lists based on start and end time
        if len(gtiffs) > 1:
            gtiffs = self._filter_img_list(
                gtiffs, start_time, end_time, self.base_tformat, self.base_tz
            )
        if len(pngs) > 1:
            pngs = self._filter_img_list(
                pngs, start_time, end_time, self.base_tformat, self.base_tz
            )

        if gtiffs and pngs:
            print(
                f"{Fore.RED} WARNING: Multiple image formats found in base image folder. Using "
                f"geotiffs..."
            )
            self.base_is_geotiff = True
            for file in gtiffs:
                fname = editortools.copy_image(file, self.basepath)
                self.baseimgfiles.append(fname)
        elif gtiffs:
            self.base_is_geotiff = True
            for file in gtiffs:
                fname = editortools.copy_image(file, self.basepath)
                self.baseimgfiles.append(fname)
        elif pngs:
            self.base_is_geotiff = False
            for file in pngs:
                fname = editortools.copy_image(file, self.basepath)
                self.baseimgfiles.append(fname)
        else:
            print(f"{Fore.RED} ERROR: No valid image formats found. Exiting...")
            sys.exit(0)

        if len(self.baseimgfiles) == 1:
            self.base_is_static = True
        else:
            self.base_is_static = False

        err2 = (
            f"{Fore.RED} ERROR: Dynamic base image requires timezone and/or"
            f"filename format to be specified. Exiting..."
        )

        if not self.base_is_static and self.base_tformat is None:
            print(err2)
            sys.exit(0)

        return

    def _setup_overlay_imgs(
        self, overlaypaths, overlay_timezones, start_time, end_time
    ):
        if overlay_timezones:
            self.overlays_tz = overlay_timezones.pop(
                "timezone", [pytz.utc] * len(overlaypaths)
            )
            overlays_format = overlay_timezones.pop(
                "filename format", [None] * len(overlaypaths)
            )
            if all(overlays_format):
                self.overlay_tformats = [
                    gistools.time_format_str(x) for x in overlays_format
                ]
            else:
                self.overlay_tformats = [
                    gistools.time_format_str("YYYY-MM-DD hh_mm_ss")
                    for x in overlays_format
                ]
        else:
            self.overlays_tz = [pytz.utc] * len(overlaypaths)
            self.overlay_tformats = [
                gistools.time_format_str("YYYY-MM-DD hh_mm_ss")
            ] * len(overlaypaths)

        err = (
            f"{Fore.RED} ERROR: Using the `overlay_timezones` parameter requires all or none of the "
            f"timezones and/or filename formats to be specified. Exiting..."
        )

        if len(self.overlays_tz) != len(overlaypaths) or len(
            self.overlay_tformats
        ) != len(overlaypaths):
            print()
            print(err)
            sys.exit(0)

        # Copy overlay files - Overlays can only be geotiff format
        self.overlayimgfiles = []
        for i, p in enumerate(overlaypaths):
            p = Path(p)
            gtiffs = list(p.glob("*.tif")) + list(p.glob("*.tiff"))
            gtiffs = sorted(gtiffs, key=lambda item: item.name)
            if not gtiffs:
                print(
                    f"{Fore.RED} ERROR: No geotiffs found in overlay path {str(p)}. Exiting..."
                )
                sys.exit(0)
            gtiffs = self._filter_img_list(
                gtiffs,
                start_time,
                end_time,
                self.overlay_tformats[i],
                self.overlays_tz[i],
            )
            row = []
            for file in gtiffs:
                fname = editortools.copy_image(file, self.overlaypaths[i])
                row.append(fname)
            self.overlayimgfiles.append(row)

        return

    def _setup_matched_lists(self):
        self.overlay_start = []
        self.overlay_end = []
        for i, overlay in enumerate(self.overlayimgfiles):
            self.overlay_start.append(
                gistools.get_timestamp(
                    overlay[0].stem, self.overlay_tformats[i], self.overlays_tz[i]
                )
            )
        for i, overlay in enumerate(self.overlayimgfiles):
            self.overlay_end.append(
                gistools.get_timestamp(
                    overlay[-1].stem, self.overlay_tformats[i], self.overlays_tz[i]
                )
            )
        self.overlay_deltas = [
            (tend - tstart).total_seconds()
            for tend, tstart in zip(self.overlay_end, self.overlay_start)
        ]

        self.trange_norms = [(np.linspace(1, x, 1) / x)
                             for x in self.overlay_deltas]

        # Calculate interval between images in each overlay path
        overlay_intervals = []
        for i, overlay in enumerate(self.overlayimgfiles):
            overlay_intervals.append(
                self._calculate_interval(
                    (overlay[0], overlay[1]),
                    (self.overlay_tformats[i],),
                    (self.overlays_tz[i],),
                )
            )

        # Calculate interval between base images, if applicable
        base_interval = 0
        if not self.base_is_static:
            base_interval = self._calculate_interval(
                (self.baseimgfiles[0], self.baseimgfiles[1]),
                (self.base_tformat,),
                (self.base_tz,),
            )
            all_intervals = overlay_intervals + [base_interval]
        else:
            all_intervals = overlay_intervals

        mindt_idx = all_intervals.index(min(all_intervals))
        self.base_is_reference = False
        self.overlay_reference_idx = mindt_idx

        if mindt_idx == len(all_intervals) - 1:
            if not self.base_is_static:
                self.framecount = len(self.baseimgfiles)
                reffiles = self.baseimgfiles
                refidx = 0
                self.base_is_reference = True
            else:
                self.framecount = len(self.overlayimgfiles[-1])
                reffiles = self.overlayimgfiles[-1]
                refidx = -1
        else:
            if not self.base_is_static:
                self.framecount = len(self.overlayimgfiles[mindt_idx])
                reffiles = self.overlayimgfiles[mindt_idx]
                refidx = mindt_idx
            else:
                self.framecount = len(self.overlayimgfiles[mindt_idx])
                reffiles = self.overlayimgfiles[mindt_idx]
                refidx = mindt_idx

        new_overlay_list = []
        for j, overlay in enumerate(self.overlayimgfiles):
            tmp_list = []
            if len(overlay) < self.framecount:
                for i, file in enumerate(reffiles):
                    if self.base_is_reference:
                        closest_idx = self._get_closest_time(
                            (file.stem, self.base_tformat),
                            ([x.stem for x in overlay], self.overlay_tformats[j]),
                            tz=(self.base_tz, self.overlays_tz[j]),
                        )
                    else:
                        closest_idx = self._get_closest_time(
                            (file.stem, self.overlay_tformats[refidx]),
                            ([x.stem for x in overlay], self.overlay_tformats[j]),
                            tz=(self.overlays_tz[refidx], self.overlays_tz[j]),
                        )

                    tmp_list.append(overlay[closest_idx])

                new_overlay_list.append(tmp_list)
            else:
                new_overlay_list.append(overlay)

        self.overlayimgfiles = new_overlay_list

        # Fill in the base file list with duplicates if needed
        if not self.base_is_static:
            if len(self.baseimgfiles) < self.framecount:
                new_base_list = []
                for i, file in enumerate(self.baseimgfiles):
                    closest_idx = self._get_closest_time(
                        (file.stem, self.base_tformat),
                        ([x.stem for x in reffiles],
                         self.overlay_tformats[refidx]),
                        tz=(self.base_tz, self.overlays_tz[refidx]),
                    )
                    new_base_list.append(self.baseimgfiles[closest_idx])
                self.baseimgfiles = new_base_list
        return

    def _get_time(self, mode="base"):
        if mode == "base":
            _times = []
            for f in self.baseimgfiles:
                _times.append(
                    gistools.get_timestamp(
                        f, self.base_tformat, tz=self.base_tz)
                )
            _times.sort()
            start = _times[0]
            end = _times[-1]
        elif mode == "overlay":
            _start = []
            _end = []
            _times = []
            for i, overlay in enumerate(self.overlayimgfiles):
                for f in overlay:
                    _times.append(
                        gistools.get_timestamp(
                            f, self.overlay_tformats[i], tz=self.overlays_tz[i]
                        )
                    )
                _times.sort()
                _start.append(_times[0])
                _end.append(_times[-1])
            start = max(_start)
            end = min(_end)

        else:
            start = None
            end = None

        return start.isoformat(), end.isoformat()

    @staticmethod
    def _bbox_error_msg(self, coord, index=None):
        if not index:
            err = (
                f"{Fore.RED} WARNING! Provided bbox {coord} extent is beyond the bounds of the base "
                f"image. Using image bounds instead."
            )
        else:
            err = (
                f"{Fore.RED} WARNING! Provided bbox {coord} extent is beyond the bounds of the "
                f"overlay image at index {str(index)}. Using image bounds instead."
            )
        return err

    @staticmethod
    def _calculate_interval(files, formatdicts, tz=None):
        """
        Calculate interval in seconds between two files
        @param files: (tup) tuple of filenames containing datetime strings
                      (e.g. '2023-01-01 10_43_00')
        @param formatdicts: (tup) tuple of dicts containing indices returned by _time_format_str
        @param tz: (tup) tuple of pytz timezone objects
        @return: interval between two files in seconds
        """
        # Extract datetime from string
        file1 = files[0].stem
        file2 = files[1].stem

        # Unpack formatdicts argument
        tformat1 = formatdicts[0]
        try:
            tformat2 = formatdicts[1]
        except IndexError:
            tformat2 = tformat1

        # Unpack tz argument
        if tz:
            tz1 = tz[0]
            try:
                tz2 = tz[1]
            except IndexError:
                tz2 = tz1
        else:
            tz1 = pytz.utc
            tz2 = pytz.utc

        t1 = gistools.get_timestamp(file1, tformat1, tz=tz1)
        t2 = gistools.get_timestamp(file2, tformat2, tz=tz2)

        # Convert times to UTC
        t1 = t1.astimezone(pytz.utc)
        t2 = t2.astimezone(pytz.utc)

        # Get delta
        delta = abs((t2 - t1).total_seconds())

        return delta

    @staticmethod
    def _filter_img_list(imglist, start_time, end_time, tformatdict, tz):
        """
        Filter a list of image filenames between start and end times
        """

        # Convert start and end times to UTC
        try:
            tstart = datetime.fromisoformat(start_time)
            if tstart.tzinfo is None:
                tstart = tz.localize(tstart)
                tstart = tstart.astimezone(pytz.utc)
            else:
                tstart = tstart.astimezone(pytz.utc)

        except TypeError:
            tstart = None

        try:
            tend = datetime.fromisoformat(end_time)
            if tend.tzinfo is None:
                tend = tz.localize(tend)
                tend = tend.astimezone(pytz.utc)
            else:
                tend = tend.astimezone(pytz.utc)

        except TypeError:
            tend = None

        new_list = []
        if imglist:
            for img in imglist:
                t = gistools.get_timestamp(img.stem, tformatdict, tz)
                t = t.astimezone(pytz.utc)
                if tend and not tstart:
                    if t <= tend:
                        new_list.append(img)
                elif tstart and not tend:
                    if t >= tstart:
                        new_list.append(img)
                elif tstart and tend:
                    if tstart <= t <= tend:
                        new_list.append(img)
                else:
                    new_list.append(img)

        return new_list

    @staticmethod
    def _get_closest_time(target, srcarray, tz=None):
        if tz:
            targettz = tz[0]
            try:
                srctz = tz[1]
            except IndexError:
                srctz = targettz
        else:
            targettz = pytz.utc
            srctz = pytz.utc

        _srcarray = srcarray[0]
        src_tformat = srcarray[1]

        target_time = gistools.get_timestamp(target[0], target[1], targettz)
        src_times = [gistools.get_timestamp(
            x, src_tformat, srctz) for x in _srcarray]

        if targettz != srctz:
            target_time = target_time.astimezone(srctz)

        closest_time = min(src_times, key=lambda d: abs(d - target_time))
        idx = src_times.index(closest_time)
        # file = srcarray[idx]

        return idx

    @staticmethod
    def _get_separator(instr, startidx, startchar):
        right_sep = None
        left_sep = None
        if startidx == 0:
            c = 1
            left_sep = None
            while True:
                try:
                    _s = instr[startidx + c]
                except KeyError:
                    right_sep = None
                    break
                if _s != startchar:
                    right_sep = _s
                    break
                c += 1
        else:
            c = 0
            while True:
                try:
                    _s = instr[startidx - c]
                except KeyError:
                    left_sep = None
                    break
                if _s != startchar:
                    left_sep = _s
                    break
                c += 1
            c = 0
            while True:
                try:
                    _s = instr[startidx + c]
                except KeyError:
                    right_sep = None
                    break
                if _s != startchar:
                    right_sep = _s
                    break
                c += 1

        return left_sep, right_sep


def create_video(vidsavepath, finaloverlaypath, fps=20, codec="rawvideo", **kwargs):
    img_filenames = list(Path(finaloverlaypath).glob("*.png"))
    img_filenames = sorted(img_filenames, key=lambda item: item.name)

    tmpdir = tempfile.TemporaryDirectory()
    tmppath = Path(tmpdir.name)
    newfiles = []

    if kwargs:
        timestamps = kwargs.get("timestamps", None)
        if timestamps:
            for f in img_filenames:
                fnew = str(tmppath / (f.stem + ".png"))
                newfiles.append(fnew)
                fmod = f.stem.replace("_", ":")
                flist = fmod.split(" ")
                timestamps["label"] = (" ").join(flist[0:2])
                kwargs["timestamps"] = timestamps
                img = Image.open(str(f))
                img = editortools.modify_image(img, **kwargs)
                if codec == "rawvideo" or codec == "png":
                    img = editortools.convert_color(img)
                img.save(fnew)
                img.close()
        else:
            for f in img_filenames:
                fnew = str(tmppath / (f.stem + ".png"))
                newfiles.append(fnew)
                img = Image.open(str(f))
                img = editortools.modify_image(img, **kwargs)
                if codec == "rawvideo" or codec == "png":
                    img = editortools.convert_color(img)
                img.save(fnew)
                img.close()

    imgarr = [str(f) for f in newfiles]
    clip = ImageSequenceClip(imgarr, fps=fps)
    clip.write_videofile(vidsavepath, codec=codec)
    shutil.rmtree(str(tmppath))

    return



from pathlib import Path
import pytz
from rasterio.enums import Resampling
from goesvideo.addons import sceneeditor

"""
This demo relies on images composited in 'create_video_goes_west_demo.py', which generates a scene of the
Hawaiian Islands using GOES-West imagery. 
"""

#---- Demo options
preview = False           # Generate a single preview image for each edit function
test_annotations = True   # Add annotations to image(s)
test_timestamps = True    # Add timestamps to image(s)
test_recrop = False       # Recrop image(s) to smaller region
test_resize = False       # Resize image(s)
test_reproject = False    # Reproject image(s) to EPSG:4326
test_tovideo = True       # Create video output

#----Setup editor
basedir = str(Path.home() / 'goesbase')
scene = 'Hawaii_Sample_Scene'
session_name = 'Hawaii_Sample_Scene_Edit_4'
editor = sceneeditor.GoesSceneEditor(basedir, scene, session_name='Hawaii_Sample_Scene_Edit_1')
fontpath = str(Path.home()) + '/.local/share/fonts/meslolgldz-nerd-font/MesloLGSDZNerdFont-Bold.ttf'
editor.set_font(fontpath, (255, 0, 1), 30)

if test_recrop:
    new_bbox = ('156d45m23s W', '20d29m22s N', '155d51m25s W', '21d08m50s N')
    editor.recrop(new_bbox, preview=preview)

if test_resize:
    editor.resize(1024, 760, resample=Resampling.nearest, preview=preview)

if test_reproject:
    editor.reproject("EPSG:4326", preview=preview)

if test_timestamps:
    editor.set_font(fontsize=60)
    editor.add_timestamps('upper-left',
                          pytz.timezone('US/Hawaii'), 'HDT',
                          preview=preview)

if test_annotations:
    if test_recrop:
      circleopts = {'radius': 7,
                    'fill': (0, 0, 255),
                    'outline': (0, 0, 0),
                    'width': 2}
      editor.add_annotation('circle', ('20d46m09s N', '156d15m15s W'),
                            label="Maui",
                            labelopts={'padding': (5, 5)},
                            **circleopts)
    else:
      editor.set_font(fontsize=50)
      circleopts = {'radius': 7,
                    'fill': (255, 0, 0),
                    'outline': (0, 0, 0),
                    'width': 2
                   }
      arrowopts = {'tiplength': 0.25,
                   'width': 3,
                   'color': (255, 0, 0)
                   }
      editor.add_annotation('circle', ('19d40m05s N', '155d27m38s W'),
                            label="Hawaii",
                            labelopts={'padding': (10, 10)},
                            **circleopts)
      editor.add_annotation('arrow', [('20d58m26s N', '158d03m37s W'), ('21d18m51s N', '157d52m02s W')],
                            label="Honolulu",
                            labelopts={'padding': (3, 3)},
                            **arrowopts)
      editor.add_annotation('text', 'upper-center',
                            label='Hawaiian Islands - GOES-West',
                            labelopts={'fontsize': 80, 'rotation': 0})

    editor.process_annotations(preview=preview)

if test_tovideo:
  editor.to_video(codec='rawvideo', fps=20)










import os
import shutil
import tempfile
from importlib.resources import files as importfiles
from pathlib import Path

import matplotlib.cm as cm
import pytz
from PIL import Image

from goesvideo.addons.overlays import GenericOverlay


def test_generic_overlays():
    # Toggle display of output images
    # Note that this will open a large number of
    # images for viewing
    show_images = False

    # Create temp folders
    tmpfolder = tempfile.TemporaryDirectory()
    tmppath = Path(tmpfolder.name)
    subpath = tmppath / "Top_Level"
    basepath = tmppath / "Base_Images"
    overlaypath = tmppath / "Overlay_Images_1"
    outputpath = tmppath / "Final_Output"
    gtiffpath = tmppath / "Gtiff_Output"
    basepath.mkdir(exist_ok=True)
    overlaypath.mkdir(exist_ok=True)
    outputpath.mkdir(exist_ok=True)
    gtiffpath.mkdir(exist_ok=True)

    # Copy test images to tmpfolders
    subpath = importfiles("goesvideo") / "tests" / "Test Images"

    # Sort and move images to correct folders
    basefiles = subpath.glob("*.png")
    overlayfiles = subpath.glob("*.tif")
    for f in basefiles:
        shutil.copy(f, str(basepath))
    for f in overlayfiles:
        shutil.copy(f, str(overlaypath))

    # Create overlay objects
    ol_timezone = {"timezone": [pytz.utc], "filename format": ["YYYY-MM-DD hh_mm_ss"]}
    base_timezone = {"timezone": pytz.utc, "filename format": "YYYY-MM-DD hh_mm_ss"}
    ol1 = GenericOverlay(
        str(basepath),
        [str(overlaypath)],
        base_timezone=base_timezone,
        overlay_timezones=ol_timezone,
    )
    ol2 = GenericOverlay(
        str(basepath),
        [str(overlaypath)],
        start_time="2024-10-08 17:19:00",
        end_time="2024-10-08 19:50:51",
    )

    # Generate overlays
    # Test 1 - Change CRS
    out_crs = "EPSG:4326"
    ol1.create_overlays(str(outputpath / "test1a"), out_crs=out_crs)
    ol2.create_overlays(str(outputpath / "test1b"), out_crs=out_crs)

    # Test 2 - Fully overlapped bbox
    bbox = [-90, 27, -82, 38]
    ol1.create_overlays(str(outputpath / "test2a"), bbox=bbox)
    ol2.create_overlays(str(outputpath / "test2b"), bbox=bbox)

    # Test 3 - Partially overlapped bbox
    bbox = [-105, 21, -90, 40]
    ol1.create_overlays(str(outputpath / "test3a"), bbox=bbox)
    ol2.create_overlays(str(outputpath / "test3b"), bbox=bbox)

    # Test 4 - Non-overlapping bbox
    bbox = [-110, -60, -100, -40]
    ol1.create_overlays(str(outputpath / "test4a"), bbox=bbox)
    ol2.create_overlays(str(outputpath / "test4b"), bbox=bbox)

    # Test 5 - Set overlay opacity to 0.3
    opacities = [0.3]
    ol1.create_overlays(str(outputpath / "test5a"), overlay_opacities=opacities)
    ol2.create_overlays(str(outputpath / "test5b"), overlay_opacities=opacities)

    # Test 6 - Save overlay images as geotiffs
    ol1.create_overlays(
        str(outputpath / "test6a"), save_overlays_gtiff=str(gtiffpath / "test6a")
    )
    ol2.create_overlays(
        str(outputpath / "test6b"), save_overlays_gtiff=str(gtiffpath / "test6b")
    )

    # Test 7 - Cumulative overlay
    ol1.create_overlays(str(outputpath / "test7a"), cumulative=True)
    ol2.create_overlays(str(outputpath / "test7b"), cumulative=True)

    # Test 8 - Cumulative overlay with cumulative colormap
    cmap = cm.get_cmap("Spectral")
    ol1.create_overlays(
        str(outputpath / "test8a"), cumulative=True, cumulative_colormap=cmap
    )
    ol2.create_overlays(
        str(outputpath / "test8b"), cumulative=True, cumulative_colormap=cmap
    )

    # Assertions to check files exist in every subdirectory
    subpaths = [
        "test1a",
        "test1b",
        "test2a",
        "test2b",
        "test3a",
        "test3b",
        "test4a",
        "test4b",
        "test5a",
        "test5b",
        "test6a",
        "test6b",
        "test7a",
        "test7b",
        "test8a",
        "test8b",
    ]

    for sp in subpaths:
        assert any((outputpath / sp).iterdir())

    # Show images if toggled on
    if show_images:
        for sp in subpaths:
            pngfile = list((outputpath / sp).glob("*.png"))
            tiffile = list((outputpath / sp).glob("*.tif"))
            if pngfile:
                Image.open(pngfile[0]).show()
            if tiffile:
                os.replace(tiffile[0], str(tiffile[0]).split(".")[0] + ".tiff")
                tf = list((outputpath / sp).glob("*.tiff"))
                Image.open(tf[0]).show()



import sys
from importlib.resources import files as importfiles

import pytest
from PIL import Image
from moviepy.video.VideoClip import VideoClip

from goesvideo.utils import editortools


def test_video_edit():
    # Toggle display of output image
    show_image = True

    # Open test video
    clip = editortools.GoesClip(
        str(importfiles("goesvideo") / "tests" / "Test Videos" / "video.mp4")
    )

    fontpath = str(importfiles("goesvideo") / "tests" / "Fonts" / "Roboto-Regular.ttf")

    # Test video annotation
    txt = {
        "label": "this is a test label",
        "position": (75, 75),
        "fontpath": fontpath,
        "fontcolor": (255, 0, 0),
        "fontsize": 5,
        "opacity": 0.7,
    }
    kwargs = {}
    kwargs["text"] = txt
    newclip = clip.annotate(0, 1, freeze=True, **kwargs)

    assert isinstance(newclip, VideoClip)

    # Test preview
    img = clip.preview(tmark=2, display=False, **kwargs)

    assert isinstance(img, Image.Image)
    if show_image:
        img.show()

    return


if __name__ == "__main__":
    sys.exit(pytest.main())
